import os

import json
from typing import Optional, Iterable
from rich.console import Console
from rich.table import Table
import threading
from rich import print
from prompt_toolkit.completion import Completer, Completion, WordCompleter
from prompt_toolkit.buffer import CompletionState
from prompt_toolkit.keys import Keys
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit import prompt

from shimoku.execution_logger import log_error

import logging

logger = logging.getLogger(__name__)

home_directory = os.path.expanduser("~")
SHIMOKU_PATH = os.path.join(home_directory, ".shimoku")
if not os.path.exists(SHIMOKU_PATH):
    os.makedirs(SHIMOKU_PATH)

CONFIG_PATH = os.path.join(SHIMOKU_PATH, "config.json")
if not os.path.exists(CONFIG_PATH):
    with open(CONFIG_PATH, "w") as f:
        f.write("")

INTERACTIVE_HISTORY_PATH = os.path.join(SHIMOKU_PATH, "interactive_history.txt")
if not os.path.exists(INTERACTIVE_HISTORY_PATH):
    with open(INTERACTIVE_HISTORY_PATH, "w") as f:
        f.write("")


environment_v = "ENVIRONMENT"
access_token_v = "ACCESS_TOKEN"
universe_id_v = "UNIVERSE_ID"
workspace_id_v = "WORKSPACE_ID"


def get_config_profiles() -> list[str]:
    """Function to get the configuration profiles
    :return: List of configuration profiles
    """
    config_file = open(CONFIG_PATH, "r")
    try:
        all_configs = json.loads(config_file.read()).get("profiles", {})
        profiles = list(all_configs.keys())
    except json.decoder.JSONDecodeError:
        profiles = []
    config_file.close()
    return profiles


def get_current_profile() -> str:
    """Function to get the current configuration profile
    :return: Current configuration profile
    """
    config_file = open(CONFIG_PATH, "r")
    try:
        all_configs = json.loads(config_file.read())
        current_profile = all_configs.get("current_profile", "default")
    except json.decoder.JSONDecodeError:
        current_profile = "default"
    config_file.close()
    return current_profile


def get_profile_config(profile: Optional[str] = None) -> dict:
    """Function to get the configuration from the config file
    :return: Configuration dictionary
    """
    config_file = open(CONFIG_PATH, "r")
    try:
        all_configs = json.loads(config_file.read())
        current_profile = profile or all_configs.get("current_profile", "default")
        if "profiles" not in all_configs:
            raise json.decoder.JSONDecodeError
        for config in all_configs["profiles"].values():
            if any(
                key not in config
                for key in [
                    environment_v,
                    access_token_v,
                    universe_id_v,
                    workspace_id_v,
                ]
            ):
                raise json.decoder.JSONDecodeError
        config = all_configs["profiles"][current_profile]
    except json.decoder.JSONDecodeError:
        config = {
            environment_v: "production",
            access_token_v: "local",
            universe_id_v: "local",
            workspace_id_v: "local",
        }
    config_file.close()
    return config


def get_all_configs() -> dict:
    """Function to get all the configurations from the config file
    :return: Configuration dictionary
    """
    config_file = open(CONFIG_PATH, "r")
    try:
        all_configs = json.loads(config_file.read())
    except json.decoder.JSONDecodeError:
        all_configs = {}
    config_file.close()
    return all_configs


def filter_dict_list(
    l: list[dict],
    field: str,
    prefix: Optional[str] = None,
    contains: Optional[str] = None,
    case_sensitive: Optional[bool] = False,
    sort_field: Optional[str] = None,
    equal_to: Optional[str] = None,
    not_equal_to: Optional[str] = None,
) -> list[dict]:
    """Function to filter a list of dictionaries
    :param l: List of dictionaries
    :param field: Field to filter
    :param prefix: Prefix to filter
    :param contains: String to filter
    :param case_sensitive: Flag to set the case sensitivity of the contains filter
    :param sort_field: Field to sort the list
    :param equal_to: String to match the items
    :param not_equal_to: String to not match the items
    :return: Filtered list of dictionaries
    """
    if not sort_field:
        sort_field = field
    if not case_sensitive:
        prefix = prefix.lower() if prefix else None
        contains = contains.lower() if contains else None
    filtered_list = []
    for element in l:
        if not isinstance(element[field], dict) and not isinstance(
            element[field], list
        ):
            value = str(element[field])
        else:
            value = json.dumps(element[field], indent=2)
        if equal_to and value != equal_to:
            continue
        if not_equal_to and value == not_equal_to:
            continue
        value = value if case_sensitive else value.lower()
        if prefix and not value.startswith(prefix):
            continue
        if contains and contains not in value:
            continue
        filtered_list.append(element)

    try:
        non_none_items = [
            x[sort_field] for x in filtered_list if x[sort_field] is not None
        ]
        if len(non_none_items) == 0:
            print(f"Can't sort the list by {sort_field} because all the items are None")
            return filtered_list
        list_type = type(non_none_items[0])
        for item in filtered_list:
            if item[sort_field] is None:
                item[sort_field] = list_type()
        return sorted(
            filtered_list,
            key=lambda x: x[sort_field]
            if not isinstance(x[sort_field], dict)
            else str(x[sort_field]),
        )
    except TypeError:
        print(f"Error sorting the list by {sort_field}")
        return filtered_list


def display_dict_list(
    items: list[dict],
    title: str,
    fields: Optional[list[str]] = None,
    table_lines: bool = False,
):
    """Function to display a list of dictionaries
    :param items: List of dictionaries
    :param title: Title of the table
    :param fields: Fields to display
    :param table_lines: Flag to set the table lines
    """
    if fields is None:
        fields = list(items[0].keys())
    table = Table(title=title, show_lines=table_lines)
    for field in fields:
        table.add_column(field)
    for item in items:
        row = []
        for field in fields:
            if not isinstance(item[field], dict) and not isinstance(item[field], list):
                row.append(str(item[field]))
            else:
                row.append(json.dumps(item[field], indent=2))
        table.add_row(*row)

    console = Console()
    print()
    console.print(table)
    print()


def list_filtering_and_display(
    items: list[dict],
    table_title: str,
    fields: Optional[list[str]] = None,
    table_lines: bool = True,
    **kwargs,
):
    """Function to filter and display a list of items
    :param items: List of items to filter and display
    :param table_title: Title of the table
    :param fields: Fields to display
    :param table_lines: Flag to set the table lines
    """
    if not items:
        print("\nNo items to list\n")
        return

    if fields is None:
        fields = list(items[0].keys())
    else:
        fields = [field for field in fields if field in items[0].keys()]

    if not fields:
        print("\nNo fields to display\n")
        return

    if not kwargs["filter_field"]:
        kwargs["filter_field"] = fields[0]
    if not kwargs["sort_field"]:
        kwargs["sort_field"] = kwargs["filter_field"]

    items = filter_dict_list(
        items,
        field=kwargs["filter_field"],
        prefix=kwargs["prefix"],
        contains=kwargs["contains"],
        case_sensitive=kwargs["case_sensitive"],
        sort_field=kwargs["sort_field"],
        equal_to=kwargs["equal_to"],
        not_equal_to=kwargs["not_equal_to"],
    )

    if kwargs["count"]:
        print(f"# of {table_title}: {len(items)}")
        return
    else:
        display_dict_list(
            items,
            title=table_title,
            fields=fields,
            table_lines=table_lines,
        )


def display_dict(item: dict, fields: list[str]):
    """Function to display a dictionary
    :param item: Dictionary to display
    :param fields: Fields to display
    """
    resource_dict = {k: v for k, v in item.items() if k in fields}
    print(json.dumps(resource_dict, indent=4))


def display_list(
    title: str,
    items: list[str],
):
    """Function to display a list
    :param title: Title of the list
    :param items: List to display
    """
    if not items:
        print("\nNo items to list\n")
        return
    print(f"\n{title}:")
    for item in items:
        print("  · " + item)
    print()


def choose_from_menu(
    options: list[str],
    title: str = "Choose an option: ",
) -> Optional[str]:
    """
    Function to choose from a list of options
    :param options: List of options to choose from
    :param title: Title of the menu
    """
    result = [None]

    def selector():
        try:
            completer = WordCompleter(options)
            selected = prompt(title, completer=completer)
            while selected not in options:
                print("Invalid option, use the TAB key to see the available options")
                selected = prompt(title, completer=completer)
            result[0] = selected
        except KeyboardInterrupt:
            pass

    thread = threading.Thread(target=selector)
    thread.start()
    thread.join()
    if not result[0]:
        log_error(logger, "No option chosen", FileNotFoundError)

    return result[0]


class PathCompleter(Completer):
    """Class to complete paths"""

    def get_completions(self, document, complete_event) -> Iterable[Completion]:
        text = document.text_before_cursor.strip()
        current_path = [".", *text.split(os.sep)]
        str_current_path = str(os.path.join(*current_path))
        offset = 0
        last_val = ""
        while not os.path.exists(str_current_path):
            last_val = current_path.pop()
            offset += len(last_val)
            if len(current_path) == 0:
                log_error(logger, "Invalid path", FileNotFoundError)
            str_current_path = str(os.path.join(*current_path))
        if os.path.isdir(str_current_path):
            for arg in os.listdir(str_current_path):
                if not arg.startswith(last_val):
                    continue
                yield Completion(
                    f"{arg}{os.sep}"
                    if os.path.isdir(os.path.join(str_current_path, arg))
                    else arg,
                    -offset,
                )


def tab_main_navigation_key_binding(add_space: bool = False):
    key_bindings = KeyBindings()

    @key_bindings.add(Keys.Tab)
    def _(event):
        """
        Custom Tab key handler: Insert the current completion, trigger completions,
        or auto-complete if only one completion is available.
        """
        buffer = event.app.current_buffer
        current_text = buffer.document.text_before_cursor.rstrip()
        buffer.delete_before_cursor(
            count=max(len(current_text) - len(current_text.rstrip()), 0)
        )

        completion_state: CompletionState = buffer.complete_state

        if completion_state:
            completions = completion_state.completions
            if completion_state.current_completion:
                # If there's an active completion, apply it and add a space
                buffer.apply_completion(completion_state.current_completion)
                if add_space:
                    buffer.insert_text(" ")
            elif len(completions) == 1:
                # If there's only one completion, apply it automatically
                buffer.apply_completion(completions[0])
                if add_space:
                    buffer.insert_text(" ")
            else:
                # Otherwise, move to the next completion
                buffer.complete_next()
        else:
            # If the buffer is empty, start the completion process
            buffer.start_completion()

    return key_bindings


def chose_from_available_paths() -> Optional[str]:
    """Function to choose from the available paths
    :return: Chosen path
    """
    result = [None]

    def selector():
        try:
            completer = PathCompleter()
            selected = prompt(
                "Choose a sub-path: ",
                completer=completer,
                key_bindings=tab_main_navigation_key_binding(),
            )
            while not os.path.exists(selected):
                print("Invalid path, use the TAB key to see the available options")
                selected = prompt(
                    "Choose a sub-path: ",
                    completer=completer,
                    key_bindings=tab_main_navigation_key_binding(),
                )
            result[0] = selected
        except KeyboardInterrupt:
            pass

    thread = threading.Thread(target=selector)
    thread.start()
    thread.join()
    if not result[0]:
        log_error(logger, "No sub-path chosen", FileNotFoundError)

    return result[0]


def input_list() -> list[str]:
    """Function to input a list of strings
    :return: List of strings
    """
    print("Enter the list of strings, one per line, press Enter twice to finish")
    items = []
    while True:
        item = input()
        if item:
            items.append(item)
        else:
            break
    return items


def input_dict() -> dict[str, str]:
    """Function to input a dictionary of strings
    :return: Dictionary of strings
    """
    print(
        "Enter the dictionary of strings, first the key and then the value, press Enter in the key step to finish"
    )
    items = {}
    while True:
        key = input("Key: ")
        if key:
            value = input("Value: ")
            items[key] = value
        else:
            break
    return items


def save_as_file(
    logger: logging.Logger,
    file_path: str,
    file_obj: bytes,
    file_name: str = "output.txt",
):
    """Function to save content to a file
    :param logger: Logger to use
    :param file_path: Path to save the file
    :param file_obj: File object to save
    :param file_name: Name of the file to save
    """
    if len(file_name.split(".")) == 1:
        file_name += ".txt"
    try:
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        file_path = os.path.join(file_path, file_name)
        with open(file_path, "wb") as write_file:
            write_file.write(file_obj)
            logger.info(f"File saved to {file_path}")
    except Exception as e:
        logger.error(f"Error saving file to {file_path}: {e}")
        return



from typing import Optional, List, Tuple

from shimoku.api.base_resource import Resource
from shimoku.api.resources.business import Business
from shimoku.api.resources.activity_template import ActivityTemplate
from shimoku.api.client import ApiClient
from shimoku.api.resources.action import Action
from shimoku.api.resources.action_template import ActionTemplate

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


class UniverseApiKey(Resource):
    _module_logger = logger
    resource_type = "apiKey"
    plural = "apiKeys"

    def __init__(
        self,
        parent: "Universe",
        uuid: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = dict(
            userType="ADMIN",
            enabled=True,
            description="",
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            params=params,
        )


class Universe(Resource):
    _module_logger = logger
    resource_type = "universe"

    def __init__(self, api_client: ApiClient, uuid: str):
        super().__init__(
            api_client=api_client,
            uuid=uuid,
            children=[Business, ActivityTemplate, Action, UniverseApiKey],
        )

    async def create_universe_api_key(self, description: str) -> UniverseApiKey:
        return await self._base_resource.create_child(
            UniverseApiKey, userType="ADMIN", enabled=True, description=description
        )

    async def get_universe_api_keys(self) -> list[UniverseApiKey]:
        return await self._base_resource.get_children(UniverseApiKey)

    async def delete_universe_api_key(self, uuid: str) -> bool:
        return await self._base_resource.delete_child(UniverseApiKey, uuid)

    # Business methods
    async def create_business(
        self, name: str, theme: Optional[dict] = None
    ) -> Business:
        if self._base_resource.api_client.playground:
            log_error(
                logger, "Cannot create business in local environment", RuntimeError
            )
        return await self._base_resource.create_child(
            Business, alias=name, theme=theme if theme else {}
        )

    async def update_business(
        self, uuid: Optional[str] = None, name: Optional[str] = None, **params
    ) -> bool:
        if params.get("new_name") is not None:
            params["name"] = params.pop("new_name")
            params["new_alias"] = True
        if self._base_resource.api_client.playground:
            uuid, name = "local", None
        return await self._base_resource.update_child(
            Business, uuid=uuid, alias=name, **params
        )

    async def get_business(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Business]:
        if self._base_resource.api_client.playground:
            uuid, name = "local", None
        return await self._base_resource.get_child(Business, uuid, name)

    async def get_businesses(self) -> List[Business]:
        return await self._base_resource.get_children(Business)

    async def delete_business(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        if self._base_resource.api_client.playground:
            log_error(logger, "Cannot delete local business", RuntimeError)
        return await self._base_resource.delete_child(Business, uuid, name)

    # Activity template methods
    async def get_activity_template(
        self, uuid: Optional[str] = None, name_version: Optional[Tuple[str, str]] = None
    ) -> Optional[ActivityTemplate]:
        return await self._base_resource.get_child(ActivityTemplate, uuid, name_version)

    @staticmethod
    def interpret_version(version: str) -> list:
        """Interpret a version string by appending '0' to each section for sorting."""
        numbers = "0123456789"
        if not version:
            return [0, 0, 0]

        def interpret_version_section(section: str) -> list:
            """Interpret a version section, considering non-numeric characters as earlier."""
            res_section = []
            number = ""
            for c in section:
                if c in numbers:
                    number += c
                else:
                    if number:
                        res_section.append(ord("z") + int(number))
                        number = ""
                    res_section.append(ord(c))
            if number:
                res_section.append(ord("z") + int(number))
            return res_section

        return [
            interpret_version_section(section + "0") for section in version.split(".")
        ]

    async def get_activity_templates(self) -> list[ActivityTemplate]:
        templates = await self._base_resource.get_children(ActivityTemplate)
        return sorted(
            templates, key=lambda t: [t["name"]] + self.interpret_version(t["version"])
        )

    # action methods
    async def get_action(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Action]:
        return await self._base_resource.get_child(Action, uuid, name)

    async def get_actions(self) -> list[Action]:
        return await self._base_resource.get_children(Action)

    async def create_action(
        self,
        name: str,
        description: Optional[str] = None,
        action_template_id: Optional[str] = None,
        universe_api_key_id: Optional[str] = None,
        libraries: Optional[List[str]] = None,
    ) -> Action:
        return await self._base_resource.create_child(
            Action,
            alias=name,
            description=description,
            actionTemplateId=action_template_id,
            universeApiKeyId=universe_api_key_id,
            pythonLibraries=libraries,
        )

    async def delete_action(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        return await self._base_resource.delete_child(Action, uuid, name)

    async def update_action(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        new_description: Optional[str] = None,
    ) -> bool:
        return await self._base_resource.update_child(
            Action,
            uuid=uuid,
            alias=name,
            new_alias=new_name,
            new_description=new_description,
        )

    # action template methods
    async def get_action_templates(self) -> list[ActionTemplate]:
        return await self._base_resource.get_children(ActionTemplate)



""""""
from typing import Optional

from io import StringIO
import pickle
import asyncio

import pandas as pd
from shimoku.api.resources.file import File
from shimoku.api.resources.app import App
from shimoku.exceptions import ShimokuFileError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging

logger = logging.getLogger(__name__)


class FilesLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the file level.
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(self, app: Optional[App]):
        self._app = app

    async def _check_for_shimoku_generated_create(self, file_name: str):
        """
        Check if a file is shimoku generated, if it is, raise an error
        :param file_name: name of the file
        """
        file: File = await self._app.get_file(name=file_name)
        if file is not None and "shimoku_generated" in file["tags"]:
            log_error(
                logger,
                f"File {file_name} is shimoku generated, you cannot overwrite it",
                ShimokuFileError,
            )

    async def _check_for_shimoku_generated_delete(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
        with_shimoku_generated: bool = False,
    ):
        """
        Check if a file is shimoku generated, if it is, raise an error
        :param file_name: name of the file
        :param with_shimoku_generated: if True, allow to delete a file with the tag 'shimoku_generated'
        """
        file: File = await self._app.get_file(uuid=uuid, name=file_name)
        if file is None:
            log_error(logger, f"File {file_name} not found", ShimokuFileError)
        if "shimoku_generated" in file["tags"] and not with_shimoku_generated:
            log_error(
                logger,
                f"File {file_name} is shimoku generated, if you are sure you want to delete it, "
                f"please set with_shimoku_generated to True",
                ShimokuFileError,
            )

    async def delete_file(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
        with_shimoku_generated: bool = False,
    ):
        """
        Delete a file
        :param uuid: uuid of the file
        :param file_name: name of the file
        :param with_shimoku_generated: if True, allow to delete a file with the tag 'shimoku_generated'
        """
        await self._check_for_shimoku_generated_delete(
            uuid, file_name, with_shimoku_generated
        )
        await self._app.delete_file(uuid=uuid, name=file_name)

    async def get_file_metadata(
        self, uuid: Optional[str] = None, file_name: Optional[str] = None
    ) -> dict:
        """
        Get file metadata
        :param uuid: uuid of the file
        :param file_name: name of the file
        :return: the metadata
        """
        return (await self._app.get_file(uuid=uuid, name=file_name)).cascade_to_dict()

    async def post_object(
        self,
        file_name: str,
        obj: bytes,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Save an object
        :param file_name: name of the file
        :param obj: object to be saved
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        """
        await self._check_for_shimoku_generated_create(file_name)
        await self._app.create_file(
            name=file_name,
            file_object=obj,
            tags=tags,
            metadata=metadata,
            overwrite=overwrite,
        )

    async def get_object(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> bytes:
        """
        Get an object
        :param file_name: name of the file
        :param uuid: uuid of the file
        :return: object
        """
        return await self._app.get_file_object(uuid=uuid, name=file_name)

    async def post_dataframe(
        self,
        file_name: str,
        df: pd.DataFrame,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Save a dataframe
        :param df: dataframe to be saved
        :param file_name: name of the file
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        :return: object
        """
        await self._check_for_shimoku_generated_create(file_name)
        dataframe_binary: bytes = df.to_csv(index=False).encode("utf-8")
        await self._app.create_file(
            name=file_name,
            file_object=dataframe_binary,
            tags=tags,
            metadata=metadata,
            overwrite=overwrite,
        )

    async def _get_dataframe(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get a dataframe file
        :param uuid: uuid of the file
        :param file_name: name of the file
        :return: dataframe
        """
        dataset_binary: bytes = await self._app.get_file_object(
            uuid=uuid, name=file_name
        )
        d = StringIO(dataset_binary.decode("utf-8"))
        df = pd.read_csv(d)

        return df.reset_index(drop=True)

    async def get_dataframe(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get a dataframe file
        :param uuid: uuid of the file
        :param file_name: name of the file
        :return: dataframe
        """
        return await self._get_dataframe(uuid=uuid, file_name=file_name)

    async def deleted_batched_dataframe(
        self, file_name: str, with_shimoku_generated: bool = False
    ):
        """
        Delete a batched dataframe
        :param file_name: name of the file
        :param with_shimoku_generated: if True, allow to delete a file with the tag 'shimoku_generated'
        """

        files = await self._app.get_files()
        files = [
            file for file in files if file["name"].startswith(file_name + "_batch_")
        ]
        if files:
            await self._check_for_shimoku_generated_delete(
                file_name=files[0]["name"],
                with_shimoku_generated=with_shimoku_generated,
            )
            logger.info(f"Deleting {len(files)} files to overwrite {file_name}")
            await asyncio.gather(
                *[self._app.delete_file(name=file["name"]) for file in files]
            )

    def post_batched_dataframe(
        self,
        file_name: str,
        df: pd.DataFrame,
        batch_size: int = 10000,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Creates an multiple files saving a dataframe in batches, for big dataframes
        Uploads a dataframe in batches, for big dataframes
        :param file_name: name of the file
        :param df: dataframe to be saved
        :param batch_size: size of the batches
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        """
        if overwrite:
            self.deleted_batched_dataframe(file_name=file_name)

        batches = [df[i : i + batch_size] for i in range(0, df.shape[0], batch_size)]
        for i, batch in enumerate(batches):
            self.post_dataframe(
                file_name=f"{file_name}_batch_{i}",
                df=batch,
                tags=tags,
                metadata=metadata,
            )

    async def get_batched_dataframe(self, file_name: str) -> pd.DataFrame:
        """
        Get a batched dataframe file group
        :param file_name: name of the file
        :return: dataframe
        """
        files = await self._app.get_files()
        files = [
            file for file in files if file["name"].startswith(file_name + "_batch_")
        ]
        files = sorted(files, key=lambda x: int(x["name"].split("_batch_")[1]))
        results = await asyncio.gather(
            *[self._get_dataframe(file["id"]) for file in files]
        )
        return pd.concat(results, ignore_index=True)

    async def post_ai_model(
        self,
        model_name: str,
        model: callable,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Save a model
        :param model_name: name of the model
        :param model: model to be saved
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        """
        await self._check_for_shimoku_generated_create(model_name)
        model_binary: bytes = pickle.dumps(model)
        await self._app.create_file(
            name=model_name,
            file_object=model_binary,
            tags=tags,
            metadata=metadata,
            overwrite=overwrite,
        )

    async def get_ai_model(
        self,
        uuid: Optional[str] = None,
        model_name: Optional[str] = None,
    ) -> any:
        """
        Get a model
        :param uuid: uuid of the file
        :param model_name: name of the model
        :return: model
        """
        model_binary: bytes = await self._app.get_file_object(
            uuid=uuid, name=model_name
        )
        return pickle.loads(model_binary)



from shimoku import Client


def speed_gauge(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "speed-gauge-test")

    shimoku_client.plt.speed_gauge(
        name="Third", value=60, min_value=0, max_value=70, order=0
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class DataSetsLayerHeader:
    """
    This class is used to interact with the API at the data set level.
    """

    def append_to_data_set(
        self,
        data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        """
        Appends data to a dataset, if the dataset does not exist it will be created.
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        :return: dataset id
        """
        pass

    def delete_data_from_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete all data from a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        pass

    def delete_data_point_from_data_set(
        self,
        data_point_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Delete a data point from a dataset in the menu path
        :param data_point_id: id of the data point
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        pass

    def delete_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        pass

    def get_columns_mapping_from_data_point(
        self,
        data: dict,
    ) -> dict:
        """Get the columns mapping from a data point in the menu path,
        assuming that the null values are empty strings.

        :param data: data to be stored in the dataset
        :return: columns mapping
        """
        pass

    def get_data_from_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> list:
        """Get the data in a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param limit: limit of the data points to be returned
        :return: data in the dataset
        """
        pass

    def get_data_set_metadata(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Get the metadata of a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :return: metadata of the dataset
        """
        pass

    def linear_update_data_from_data_set(
        self,
        data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Update the data in the dataset following a linear update strategy,
        creates the new data points if they do not exist, deletes the old ones if they do not exist in the new data,
        and linearly checks if the data points have changed and updates them.
        This makes this function more efficient for small changes than replace_data_from_data_set.
        But if the changes are linear in proportion to the size of the dataset, this function can be
        slower than replace_data_from_data_set.

        :param data: data to be stored in the dataset
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        pass

    def replace_data_from_data_set(
        self,
        data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Replace the data in a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        """
        pass

    def update_data_point_from_data_set(
        self,
        data_point_id: str,
        data: dict,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Update a data point from a dataset in the menu path,
        data has to follow the schema of the dataset!

        e.g. if your dataset has columns 'id', 'name' and 'age',
        the format to update a data point should be:
        {'stringField2': 'new_name', 'intField1': 25}
             ^ because 'id' is the first string column and you wouldn't want to update it

        suggestion: use get_columns_mapping_from_data_point
        to get the columns mapping from a data point

        :param data_point_id: id of the data point
        :param data: data to be stored in the dataset
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        pass



from typing import Optional, Union, TYPE_CHECKING
from math import floor, ceil, log10
from pandas import DataFrame
from copy import deepcopy

from shimoku.plt.utils import validate_data_is_pandarable, ShimokuPalette
from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def line_and_bar_charts(
    self: "PlotLayer",
    order: int,
    x: str,
    data: Optional[Union[str, DataFrame, list[dict]]],
    bar_names: Optional[list[str]] = None,
    line_names: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    bar_axis_name: Optional[str] = None,
    bar_suffix: Optional[str] = None,
    line_axis_name: Optional[str] = None,
    line_suffix: Optional[str] = None,
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a chart with bars and lines."""
    data_mappings_to_tuples = await self._choose_data(order, data)
    if isinstance(data, str):
        data = self._shared_data[data]

    if bar_names is None:
        bar_names = ["bar"]

    if line_names is None:
        line_names = ["line"]

    df: DataFrame = validate_data_is_pandarable(data)
    max_bars = df[bar_names].max().max()
    pow_10 = floor(log10(abs(max_bars)))
    max_bars = ceil(max_bars / 10**pow_10) * 10**pow_10
    bar_interval = abs(max_bars) / (5 if 1000 > max_bars > 0.01 else 10)

    longest_value_bars = max(len(str(max_bars)), len(str(bar_interval)))

    max_lines = df[line_names].max().max()
    pow_10 = floor(log10(abs(max_lines)))
    max_lines = ceil(max_lines / 10**pow_10) * 10**pow_10
    line_interval = abs(max_lines) / (5 if 1000 > max_lines > 0.01 else 10)

    longest_value_lines = max(len(str(max_lines)), len(str(line_interval)))

    common_options = get_common_echart_options()
    x_axis = common_options["xAxis"][0]
    del x_axis["data"]
    x_axis["type"] = "value"
    x_axis["scale"] = True
    common_options["yAxis"][0]["scale"] = True

    common_options.update(
        {
            "xAxis": [
                {
                    "data": "#set_data#",
                    "fontFamily": "Rubik",
                    "axisPointer": {"type": "shadow"},
                    "type": "category",
                    "nameLocation": "middle",
                    "nameGap": 35,
                }
            ],
            "yAxis": [
                {
                    "type": "value",
                    "fontFamily": "Rubik",
                    "axisLabel": {"formatter": "{value}" + bar_suffix},
                    "nameLocation": "middle",
                    "nameGap": 24 + 7 * (longest_value_bars + int(len(bar_suffix))),
                    "alignTicks": True,
                    "axisLine": {
                        "show": True,
                        "lineStyle": {"color": ShimokuPalette.CHART_C1.value},
                    },
                    "splitNumber": 5,
                    "interval": bar_interval,
                    "max": max_bars,
                },
                {
                    "type": "value",
                    "fontFamily": "Rubik",
                    "axisLabel": {"formatter": "{value}" + line_suffix},
                    "nameLocation": "middle",
                    "nameGap": 24 + 7 * (longest_value_lines + int(len(line_suffix))),
                    "axisLine": {
                        "show": True,
                        "lineStyle": {"color": f"var(--chart-C{len(bar_names)+2})"},
                    },
                    "splitNumber": 5,
                    "interval": line_interval,
                    "max": max_lines,
                },
            ],
        }
    )
    series_options = get_common_series_options()

    series = []
    for index, name in enumerate(bar_names):
        sop = deepcopy(series_options)
        sop.update(
            {
                "name": name,
                "type": "bar",
                "itemStyle": {
                    "color": f"var(--chart-C{index+1})",
                    "borderRadius": [9, 9, 0, 0],
                },
                "emphasis": {
                    "itemStyle": {
                        "color": f"var(--chart-C{index+1})",
                        "borderRadius": [9, 9, 0, 0],
                    },
                    "lineStyle": {
                        "color": f"var(--chart-C{index+1})",
                    },
                },
                "smooth": True,
            }
        )
        series.append(sop)

    for index, name in enumerate(line_names):
        sop = deepcopy(series_options)
        sop.update(
            {
                "name": name,
                "type": "line",
                "yAxisIndex": 1,
                "itemStyle": {
                    "color": f"var(--chart-C{len(bar_names) + index + 1})",
                    "borderRadius": [9, 9, 0, 0],
                },
                "emphasis": {
                    "itemStyle": {
                        "color": f"var(--chart-C{len(bar_names) + index + 1})",
                        "borderRadius": [9, 9, 0, 0],
                    },
                    "lineStyle": {
                        "color": f"var(--chart-C{len(bar_names) + index + 1})",
                    },
                },
                "smooth": True,
            }
        )
        series.append(sop)

    await self._create_trend_chart(
        echart_options=common_options,
        series_options=series,
        axes=x,
        values=bar_names + line_names,
        option_modifications=option_modifications,
        data_mapping_to_tuples=data_mappings_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=[bar_axis_name, line_axis_name],
        variant=variant,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



from shimoku import Client
from .data.dynamic_conditional_and_auto_send_input_form import form_groups


def dynamic_conditional_and_auto_send_input_form(shimoku_client: Client):
    shimoku_client.set_menu_path("Input forms", "dynamic conditional and auto send")

    shimoku_client.plt.generate_input_form_groups(
        order=0, form_groups=form_groups, dynamic_sequential_show=True
    )

    shimoku_client.plt.generate_input_form_groups(
        order=1,
        form_groups={
            "Personal information": form_groups["Personal information"],
            "Other data": form_groups["Other data"],
        },
        auto_send=True,
        dynamic_sequential_show=True,
    )



from typing import Union, Optional, TypedDict, TYPE_CHECKING
from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.business import Business
    from shimoku.api.resources.dashboard import Dashboard
    from shimoku.api.resources.app import App
    from shimoku.api.user_access_classes.businesses_layer import WorkspacesLayer
    from shimoku.api.user_access_classes.apps_layer import MenuPathsLayer
    from shimoku.api.user_access_classes.dashboards_layer import BoardsLayer
import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)

VALID_PERMISSIONS = ["READ", "WRITE"]
VALID_RESOURCES = ["DATA", "DATA_EXECUTION", "USER_MANAGEMENT", "BUSINESS_INFO"]
VALID_TARGETS = ["GROUP", "USER"]


class Role(Resource):
    resource_type = "role"
    alias_field = "role"
    plural = "roles"

    class RoleParams(TypedDict):
        permission: str
        resource: str
        target: str
        role: str

    def __init__(
        self,
        parent: Union["Business", "Dashboard", "App"],
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params: Role.RoleParams = dict(
            permission="READ",
            resource="BUSINESS_INFO",
            target="GROUP",
            role=alias if alias else "",
        )
        super().__init__(
            parent=parent,
            uuid=uuid,
            check_params_before_creation=["role"],
            params=params,
            db_resource=db_resource,
        )

    async def delete(self):
        return await self._base_resource.delete()

    async def update(self):
        return await self._base_resource.update()

    def set_params(
        self,
        permission: Optional[str] = None,
        resource: Optional[str] = None,
        target: Optional[str] = None,
        role: Optional[str] = None,
    ):
        if permission:
            if permission and permission not in VALID_PERMISSIONS:
                log_error(
                    logger,
                    f"{permission} is not a valid value for permission, "
                    f"the valid values are: {VALID_PERMISSIONS}",
                    ValueError,
                )
            self._base_resource.params["permission"] = permission
        if resource:
            if resource not in VALID_RESOURCES:
                log_error(
                    logger,
                    f"{resource} is not a valid value for resource, "
                    f"the valid values are: {VALID_RESOURCES}",
                    ValueError,
                )
            self._base_resource.params["resource"] = resource
        if target:
            if target and target not in VALID_TARGETS:
                log_error(
                    logger,
                    f"{target} is not a valid value for target, "
                    f"the valid values are: {VALID_TARGETS}",
                    ValueError,
                )
            self._base_resource.params["target"] = target
        if role:
            self._base_resource.params["role"] = role


# Parent class methods
async def create_role(
    self: Union["Business", "Dashboard", "App"],
    role: str,
    permission: str = "READ",
    resource: str = "BUSINESS_INFO",
    target: str = "GROUP",
) -> Role:
    return await self._base_resource.create_child(
        Role, alias=role, permission=permission, resource=resource, target=target
    )


async def get_role(
    self: Union["Business", "Dashboard", "App"],
    uuid: Optional[str] = None,
    role: Optional[str] = None,
) -> Optional[Role]:
    return await self._base_resource.get_child(Role, uuid, role)


async def get_roles(self: Union["Business", "Dashboard", "App"]) -> list[Role]:
    return await self._base_resource.get_children(Role)


async def delete_role(
    self: Union["Business", "Dashboard", "App"],
    uuid: Optional[str] = None,
    role: Optional[str] = None,
) -> Role:
    return await self._base_resource.delete_child(Role, uuid, role)


# User level methods
async def user_create_role(
    self: Union["WorkspacesLayer", "MenuPathsLayer", "BoardsLayer"],
    uuid: Optional[str] = None,
    name: Optional[str] = None,
    resource: Optional[str] = None,
    role_name: Optional[str] = None,
    permission: Optional[str] = None,
    target: Optional[str] = None,
) -> dict:
    """
    Create a new role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param resource: resource of the role
    :param role_name: name of the role
    :param permission: permission of the role
    :param target: target of the role
    """
    resource_obj = await self._get_for_roles(uuid, name)

    return (
        await resource_obj.create_role(
            role=role_name, permission=permission, resource=resource, target=target
        )
    ).cascade_to_dict()


async def user_delete_role(
    self: Union["WorkspacesLayer", "MenuPathsLayer", "BoardsLayer"],
    uuid: Optional[str] = None,
    name: Optional[str] = None,
    role_id: Optional[str] = None,
    role_name: Optional[str] = None,
):
    """
    Delete the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_id: id of the role
    :param role_name: name of the role
    """
    resource = await self._get_for_roles(uuid, name)

    if resource:
        await resource.delete_role(uuid=role_id, role=role_name)


async def user_get_role(
    self: Union["WorkspacesLayer", "MenuPathsLayer", "BoardsLayer"],
    uuid: Optional[str] = None,
    name: Optional[str] = None,
    role_name: Optional[str] = None,
) -> list[dict]:
    """
    Get the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_name: name of the role
    """
    resource = await self._get_for_roles(uuid, name)

    if resource:
        role = await resource.get_role(role=role_name)
        if role:
            return role.cascade_to_dict()
        logger.info(f"Role {role_name} not found in {str(resource)}")

    return []


async def user_get_roles(
    self: Union["WorkspacesLayer", "MenuPathsLayer", "BoardsLayer"],
    uuid: Optional[str] = None,
    name: Optional[str] = None,
) -> list[dict]:
    """
    Get the roles at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    """
    resource = await self._get_for_roles(uuid, name)

    if resource:
        return [role.cascade_to_dict() for role in await resource.get_roles()]

    return []



import collections.abc
import json
import json5
import logging
from typing import Optional, Union
from enum import Enum

import numpy as np
from pandas import DataFrame

from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


def create_normalized_name(name: str) -> str:
    """Having a name create a normalizedName

    Example
    ----------------------
    # "name": "   Test Borrar_grafico    "
    # "normalizedName": "test-borrar-grafico"
    """
    if any(
        [
            c
            not in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._- &$"
            for c in name.encode("ascii", "ignore").decode()
        ]
    ):
        log_error(
            logger,
            f'You can only use letters, numbers, spaces, "-" and "_" in your name | '
            f"you introduced {name}",
            ValueError,
        )
    # remove empty spaces at the beginning and end
    name: str = name.strip()
    # replace "_" for www protocol it is not good
    name = name.replace("_", "-")

    return "-".join(name.split(" ")).lower()


def clean_menu_path(menu_path: str) -> tuple[str, str]:
    """
    Break the menu path in the apptype or app normalizedName
    and the path normalizedName if any
    :param menu_path: the menu path
    """
    # remove empty spaces
    menu_path: str = menu_path.strip()
    # replace "_" for www protocol it is not good
    menu_path = menu_path.replace("_", "-")

    try:
        assert len(menu_path.split("/")) <= 2  # we allow only one level of path
    except AssertionError:
        raise ValueError(
            f"We only allow one subpath in your request | "
            f"you introduced {menu_path} it should be maximum "
            f'{"/".join(menu_path.split("/")[:1])}'
        )

    # Split AppType or App Normalized Name
    normalized_name: str = menu_path.split("/")[0]
    name: str = " ".join(normalized_name.split("-"))

    try:
        path_normalized_name: str = menu_path.split("/")[1]
        path_name: str = " ".join(path_normalized_name.split("-"))
    except IndexError:
        path_name = None

    return name, path_name


def calculate_percentages_from_list(numbers, round_digits_min):
    """Calculate the proportion of each number in a list
    :param numbers: list of numbers
    :param round_digits_min: minimum number of digits to round
    :return: list of percentages
    """

    def max_precision():
        max_p = 0
        for n in numbers:
            str_n = str(n)
            if "." in str_n:
                n_precision = len(str_n.split(".")[1])
                max_p = n_precision if n_precision > max_p else max_p
        return max(max_p, round_digits_min)

    if isinstance(numbers, list):
        numbers = np.array(numbers)

    perc = np.round(100 * numbers / np.sum(numbers), max_precision())
    round_max = 99.9
    while np.sum(perc) > 99.99:
        perc = np.round(round_max * numbers / np.sum(numbers), max_precision())
        round_max -= 0.1
    return perc


# From https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
def deep_update(source, overrides) -> list:
    """
    Update a nested dictionary or similar mapping.
    Modify ``source`` in place.
    """
    for key, value in overrides.items():
        if isinstance(value, collections.abc.Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
    return source


def get_uuids_from_dict(_dict: dict) -> list[str]:
    """Get all uuids from a dictionary. They follow the pattern '#{'id'}'."""
    uuids = []
    for k, v in _dict.items():
        if isinstance(v, dict):
            uuids.extend(get_uuids_from_dict(v))
        elif isinstance(v, list):
            uuids.extend(get_uuids_from_list(v))
        elif isinstance(v, str) and v.startswith("#{") and v.endswith("}"):
            uuids.append(v[2:-1])
    return uuids


def get_uuids_from_list(_list: list) -> list[str]:
    """Get all uuids from a list. They follow the pattern '#{'id'}'."""
    uuids = []
    for v in _list:
        if isinstance(v, dict):
            uuids.extend(get_uuids_from_dict(v))
        elif isinstance(v, list):
            uuids.extend(get_uuids_from_list(v))
        elif isinstance(v, str) and v.startswith("#{") and v.endswith("}"):
            uuids.append(v[2:-1])
    return uuids


def get_data_references_from_dict(
    _dict: dict, previous_keys: Optional[list[Union[str, int]]] = None
) -> list[list[str]]:
    """Get all data references from a dictionary. They follow the pattern '#set_data#'."""
    if previous_keys is None:
        previous_keys = []
    entries = []
    for k, v in _dict.items():
        if isinstance(v, dict):
            entries.extend(
                get_data_references_from_dict(v, previous_keys=previous_keys + [k])
            )
        elif isinstance(v, list):
            entries.extend(
                get_data_references_from_list(v, previous_keys=previous_keys + [k])
            )
        elif v == "#set_data#":
            entries.append(previous_keys + [k])
    return entries


def get_data_references_from_list(
    _list: list, previous_keys: Optional[list[Union[int, str]]] = None
) -> list[list[str]]:
    """Get all data references from a list. They follow the pattern '#set_data#'."""
    if previous_keys is None:
        previous_keys = []
    entries = []
    for i, v in enumerate(_list):
        if isinstance(v, dict):
            entries.extend(
                get_data_references_from_dict(v, previous_keys=previous_keys + [i])
            )
        elif isinstance(v, list):
            entries.extend(
                get_data_references_from_list(v, previous_keys=previous_keys + [i])
            )
        elif v == "#set_data#":
            entries.append(previous_keys + [i])
    return entries


def validate_data_is_pandarable(
    data: Union[str, DataFrame, list[dict], dict]
) -> DataFrame:
    """"""
    if isinstance(data, DataFrame):
        df_ = data.copy()
    elif isinstance(data, list):
        try:
            df_ = DataFrame(data)
        except Exception:
            raise ValueError(
                "The data you passed is a list that must be "
                "able to be converted into a pandas dataframe"
            )
    elif isinstance(data, dict):
        try:
            df_ = DataFrame(data)
        except Exception:
            try:
                df_ = DataFrame(data, index=[0])
            except Exception:
                raise ValueError(
                    "The data you passed is a dict that must be "
                    "able to be converted into a pandas dataframe"
                )
    elif isinstance(data, str):
        try:
            d: list[dict] = json.loads(data)
            df_ = DataFrame(d)
        except Exception:
            raise ValueError(
                "The data you passed is a json that must be "
                "able to be converted into a pandas dataframe"
            )
    else:
        raise ValueError(
            "Input data must be a pandas dataframe, " "a json or a list of dictionaries"
        )
    return df_


def validate_table_data(
    self,
    data: Union[str, DataFrame, list[dict], dict],
    elements: list[str],
):
    """"""
    df_: DataFrame = self._validate_data_is_pandarable(data)

    cols = df_.columns
    try:
        assert all([element in cols for element in elements])
    except AssertionError:
        raise ValueError(
            "Some column names you are specifying " "are not in the input dataframe"
        )

    try:
        len_df_: int = len(df_)
        assert all([len_df_ == len(df_[~df_[element].isna()]) for element in elements])
    except AssertionError:
        raise ValueError(f"Some of the variables {elements} have none values")


def validate_tree_data(
    self,
    data: Union[str, list[dict]],
    vals: list[str],
):
    """To validate Tree and Treemap data"""
    if isinstance(data, list):
        pass
    elif isinstance(data, dict):
        pass
    elif isinstance(data, str):
        data = json.loads(data)
    else:
        raise ValueError("data must be either a list, dict or a json")

    try:
        assert sorted(data.keys()) == sorted(vals)
    except AssertionError:
        raise ValueError('data keys must be "name", "value" and "children"')


def validate_input_form_data(self, data: dict):
    try:
        assert type(data) == dict
    except AssertionError:
        raise ValueError("data must be a dict")

    try:
        assert "fields" in data
    except AssertionError:
        raise ValueError('"fields" is not a key in the input data')

    try:
        assert type(data["fields"]) == list
    except AssertionError:
        raise ValueError("fields must be a list")

    try:
        assert all(["fields" in field_ for field_ in data["fields"]])
    except AssertionError:
        raise ValueError('"fields" are not keys in the input data')

    try:
        assert all(
            [
                "fieldName" in field__ and "mapping" in field__
                for field_ in data["fields"]
                for field__ in field_["fields"]
            ]
        )
    except AssertionError:
        raise ValueError('"fieldName" and "mapping" are not keys in the input data')


def is_report_data_empty(
    report_data: Union[list[dict], str, DataFrame, dict, list]
) -> bool:
    if isinstance(report_data, DataFrame):
        if report_data.empty:
            return True
        else:
            return False
    elif isinstance(report_data, list) or isinstance(report_data, dict):
        if report_data:
            return False
        else:
            return True
    elif isinstance(report_data, str):
        report_data_: list[dict] = json.loads(report_data)
        if report_data_:
            return False
        else:
            return True
    else:
        raise ValueError(
            f"Data must be a dictionary, JSON or pandas DataFrame "
            f"Provided: {type(report_data)}"
        )


def add_sorting_to_df(
    df: DataFrame, sort: Optional[dict] = None
) -> tuple[DataFrame, dict]:
    """Add sorting to the data frame. If no sorting is provided the a column named 'sort_values' is added.
    :param df: the data frame
    :param sort: the sorting
    :return: the data frame with sorting and the sorting
    """
    if "sort_values" not in df.columns:
        df["sort_values"] = range(len(df))
    sort = {"field": "sort_values", "direction": "asc"} if not sort else sort
    df = df[
        [col for col in df.columns.tolist() if col != "sort_values"] + ["sort_values"]
    ]
    return df, sort


class ShimokuPalette(Enum):
    """Enum for color variables"""

    SUCCESS = "var(--color-success)"
    SUCCESS_LIGHT = "var(--color-success-light)"
    WARNING = "var(--color-warning)"
    WARNING_LIGHT = "var(--color-warning-light)"
    ERROR = "var(--color-error)"
    ERROR_LIGHT = "var(--color-error-light)"
    STATUS_ERROR = "var(--color-status-error)"
    WHITE = "var(--color-white)"
    BLACK = "var(--color-black)"
    GRAY = "var(--color-gray)"
    BASE_ICON = "var(--color-base-icon)"
    BACKGROUND = "var(--background-default)"
    BACKGROUND_PAPER = "var(--background-paper)"
    PRIMARY = "var(--color-primary)"
    PRIMARY_LIGHT = "var(--color-primary-light)"
    PRIMARY_DARK = "var(--color-primary-dark)"
    SECONDARY = "var(--color-secondary)"
    SECONDARY_LIGHT = "var(--color-secondary-light)"
    SECONDARY_DARK = "var(--color-secondary-dark)"
    STRIPE = "var(--color-stripe)"
    STRIPE_LIGHT = "var(--color-stripe-light)"
    COM_RED = "var(--complementary-red)"
    COM_RED_LIGHT = "var(--complementary-red-light)"
    COM_GREEN = "var(--complementary-green)"
    COM_YELLOW = "var(--complementary-yellow)"
    COM_ORANGE = "var(--complementary-orange)"
    COM_AQUA = "var(--complementary-aqua)"
    COM_VIOLET = "var(--complementary-violet)"
    CHART_C1 = "var(--chart-C1)"
    CHART_C2 = "var(--chart-C2)"
    CHART_C3 = "var(--chart-C3)"
    CHART_C4 = "var(--chart-C4)"
    CHART_C5 = "var(--chart-C5)"
    CHART_C6 = "var(--chart-C6)"
    CHART_C7 = "var(--chart-C7)"
    CHART_C8 = "var(--chart-C8)"
    CHART_C9 = "var(--chart-C9)"
    CHART_C10 = "var(--chart-C10)"


def interpret_color(color_def: Union[list, str, int]) -> Union[str, dict]:
    def rbg_to_hex(r: int, g: int, b: int) -> str:
        def clamp(x: int) -> int:
            return max(0, min(x, 255))

        # from https://stackoverflow.com/questions/3380726/converting-an-rgb-color-tuple-to-a-hexidecimal-string
        return "#{0:02x}{1:02x}{2:02x}".format(clamp(r), clamp(g), clamp(b))

    # from https://xkcd.com/color/rgb/
    color_defs = {
        "purple": "#7e1e9c",
        "red": "#e50000",
        "green": "#15b01a",
        "blue": "#0343df",
        "pink": "#ff81c0",
        "brown": "#653700",
        "orange": "#f97306",
        "yellow": "#ffff14",
        "gray": "#929591",
        "violet": "#9a0eea",
        "cyan": "#00ffff",
        "success": ShimokuPalette.SUCCESS.value,
        "success-light": ShimokuPalette.SUCCESS_LIGHT.value,
        "warning": ShimokuPalette.WARNING.value,
        "warning-light": ShimokuPalette.WARNING_LIGHT.value,
        "error": ShimokuPalette.ERROR.value,
        "error-light": ShimokuPalette.ERROR_LIGHT.value,
        "status-error": ShimokuPalette.STATUS_ERROR.value,
        "white": ShimokuPalette.WHITE.value,
        "black": ShimokuPalette.BLACK.value,
        "base-icon": ShimokuPalette.BASE_ICON.value,
        "background": ShimokuPalette.BACKGROUND.value,
        "background-paper": ShimokuPalette.BACKGROUND_PAPER.value,
        "primary": ShimokuPalette.PRIMARY.value,
        "primary-light": ShimokuPalette.PRIMARY_LIGHT.value,
        "primary-dark": ShimokuPalette.PRIMARY_DARK.value,
        "main": ShimokuPalette.CHART_C1.value,
        "secondary": ShimokuPalette.SECONDARY.value,
        "secondary-light": ShimokuPalette.SECONDARY_LIGHT.value,
        "secondary-dark": ShimokuPalette.SECONDARY_DARK.value,
        "active": ShimokuPalette.CHART_C2.value,
        "caution": ShimokuPalette.CHART_C3.value,
    }

    if isinstance(color_def, int):
        return f"var(--chart-C{abs(color_def)})"
    elif isinstance(color_def, (list, tuple)):
        color_def = rbg_to_hex(color_def[0], color_def[1], color_def[2])
    elif color_def in color_defs:
        color_def = color_defs[color_def]

    return color_def


def transform_dict_js_to_py(options_str: str):
    """https://discuss.dizzycoding.com/how-to-convert-raw-javascript-object-to-python-dictionary/"""
    options_str = options_str.replace("\n", "")
    options_str = options_str.replace(";", "")
    return json5.loads(options_str)


def retrieve_data_from_options(options: dict) -> Union[dict, list]:
    """Retrieve data from eCharts options

    Example
    -----------
    input options = {'title': {'text': 'Stacked Area Chart'},
         'tooltip': {'trigger': 'axis',
          'axisPointer': {'type': 'cross', 'label': {'backgroundColor': '#6a7985'}}},
         'legend': {'data': ['Email',
           'Union Ads',
           'Video Ads',
           'Direct',
           'Search Engine']},
         'toolbox': {'feature': {'saveAsImage': {}}},
         'grid': {'left': '3%', 'right': '4%', 'bottom': '3%', 'containLabel': True},
         'xAxis': [{'type': 'category',
           'boundaryGap': False,
           'data': ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']}],
         'yAxis': [{'type': 'value'}],
         'series': [{'name': 'Email',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [120, 132, 101, 134, 90, 230, 210]},
          {'name': 'Union Ads',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [220, 182, 191, 234, 290, 330, 310]},
          {'name': 'Video Ads',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [150, 232, 201, 154, 190, 330, 410]},
          {'name': 'Direct',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [320, 332, 301, 334, 390, 330, 320]},
          {'name': 'Search Engine',
           'type': 'line',
           'stack': 'Total',
           'label': {'show': True, 'position': 'top'},
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [820, 932, 901, 934, 1290, 1330, 1320]}]
        }

    output
        [{'Mon': 120,
          'Tue': 132,
          'Wed': 101,
          'Thu': 134,
          'Fri': 90,
          'Sat': 230,
          'Sun': 210},
         {'Mon': 220,
          'Tue': 182,
          'Wed': 191,
          'Thu': 234,
          'Fri': 290,
          'Sat': 330,
          'Sun': 310},
         {'Mon': 150,
          'Tue': 232,
          'Wed': 201,
          'Thu': 154,
          'Fri': 190,
          'Sat': 330,
          'Sun': 410},
         {'Mon': 320,
          'Tue': 332,
          'Wed': 301,
          'Thu': 334,
          'Fri': 390,
          'Sat': 330,
          'Sun': 320},
         {'Mon': 820,
          'Tue': 932,
          'Wed': 901,
          'Thu': 934,
          'Fri': 1290,
          'Sat': 1330,
          'Sun': 1320
        }
    ]
    """
    rows = []
    data = []
    cols = []
    if "xAxis" in options:
        if type(options["xAxis"]) == list:
            if len(options["xAxis"]) == 1:
                if "data" in options["xAxis"][0]:
                    rows = options["xAxis"][0]["data"]
        elif type(options["xAxis"]) == dict:
            if "data" in options["xAxis"]:
                rows = options["xAxis"]["data"]
            elif type(options["xAxis"]) == dict:
                if "data" in options["yAxis"]:
                    rows = options["yAxis"]["data"]
        else:
            raise ValueError("xAxis has multiple values only 1 allowed")
    elif "radar" in options:
        if "indicator" in options["radar"]:
            rows = [element["name"] for element in options["radar"]["indicator"]]
        elif type(options["radar"]) == dict:
            raise NotImplementedError("Multi-radar not implemented")

    if "data" in options:
        data = options["data"]
    if "series" in options:
        if "data" in options["series"]:
            pass
        else:
            for serie in options["series"]:
                if "data" in serie:
                    if serie.get("type") in ["pie", "gauge"]:
                        for datum in serie["data"]:
                            data.append(datum)
                    elif serie.get("type") == "radar":
                        for datum in serie["data"]:
                            data.append(datum["value"])
                            cols.append(datum["name"])
                        break
                    else:
                        data.append(serie["data"])

                if "name" in serie:
                    cols.append(serie["name"])
                elif "type" in serie:
                    cols.append(serie["type"])
    else:
        return {}

    df = DataFrame(data)
    if not rows and not cols:
        return df.reset_index().to_dict(orient="records")
    if not rows:
        return df.to_dict(orient="records")

    if rows:
        df.columns = rows
    df_ = df.T
    df_.columns = cols
    return df_.reset_index().to_dict(orient="records")



import tqdm
from typing import Optional, List

from shimoku.api.resources.business import Business

from shimoku.utils import create_function_name, create_normalized_name

from shimoku.code_gen.file_generator import CodeGenFileHandler
from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
    AppCodeGen,
)
from shimoku.code_gen.utils_code_gen import code_gen_from_list, code_gen_from_dict

from shimoku.execution_logger import ClassWithLogging
import logging

logger = logging.getLogger(__name__)


class BusinessCodeGen(ClassWithLogging):
    _module_logger = logger

    def __init__(self, business: Business, output_path: str):
        self._business = business
        self._output_path = f'{output_path}/{create_function_name(business["name"])}'
        self._file_generator = CodeGenFileHandler(self._output_path)

    async def generate_code(
        self, business_id: str, menu_paths: List[str], pbar: Optional[tqdm.tqdm] = None
    ):
        """Use the resources in the API to generate code_lines for the SDK. Create a file in
        the specified path with the generated code_lines.
        :param business_id: business id to use
        :param menu_paths: list of menu paths to generate code for
        :param pbar: progress bar to use
        """
        import_code_lines: List[str] = [
            "from shimoku import Client",
        ]
        main_code_lines: List[str] = [
            f'shimoku_client.set_workspace("{business_id}")',
            "",
        ]

        if self._business["theme"]:
            theme_code_lines = code_gen_from_dict(self._business["theme"], deep=4)
            main_code_lines.extend(
                [
                    f'shimoku_client.workspaces.update_workspace("{business_id}", theme={theme_code_lines[0][4:]})',
                    *theme_code_lines[1:],
                    "",
                ]
            )

        if menu_paths:
            menu_paths = [create_normalized_name(menu_path) for menu_path in menu_paths]
        apps = [
            app
            for app in sorted(await self._business.get_apps(), key=lambda x: x["order"])
            if menu_paths is None or app["normalizedName"] in menu_paths
        ]

        extra_apps_for_dashboard = {}
        last_dashboard = None
        dashboards = sorted(
            await self._business.get_dashboards(), key=lambda x: x["order"]
        )
        created_dashboards = []
        for app in apps:
            app_id = app["id"]
            app_code_gen = AppCodeGen(app, self._output_path, pbar=pbar)

            await app_code_gen.generate_code()

            import_code_lines.append(
                f"from .{app_code_gen.app_f_name}.app import {app_code_gen.app_f_name}"
            )
            aux_code_lines = []
            try:
                first_dashboard, *extra_dashboards = [
                    dashboard
                    for dashboard in dashboards
                    if app_id in await dashboard.list_app_ids()
                ]
            except IndexError:
                self._module_logger.warning(f'App {app["name"]} has no dashboards')
                continue
            created_dashboards.append(first_dashboard)
            if first_dashboard["name"] != last_dashboard:
                aux_code_lines.append(
                    f'shimoku_client.set_board("{first_dashboard["name"]}")'
                )
                last_dashboard = first_dashboard["name"]
            for dashboard in extra_dashboards:
                extra_apps_for_dashboard.setdefault(dashboard["name"], []).append(
                    app["name"]
                )

            aux_code_lines.append(f"{app_code_gen.app_f_name}(shimoku_client)")
            main_code_lines.extend(aux_code_lines)

        for dashboard_name, app_names in extra_apps_for_dashboard.items():
            app_names_code_lines = code_gen_from_list(app_names, deep=4)
            main_code_lines.extend(
                [
                    "",
                    "shimoku_client.boards.group_menu_paths("
                    '    name="' + dashboard_name + '",',
                    f"    menu_path_names={app_names_code_lines[0][4:]}",
                    *app_names_code_lines[1:],
                    ")",
                ]
            )

        if created_dashboards != dashboards:
            dashboards_code_lines = code_gen_from_list(
                [dashboard["name"] for dashboard in dashboards], deep=4
            )
            main_code_lines.extend(
                [
                    "",
                    "shimoku_client.workspaces.change_boards_order(",
                    f'   uuid="{business_id}",',
                    f"   boards={dashboards_code_lines[0][4:]}",
                    *dashboards_code_lines[1:],
                    ")",
                ]
            )

        business_name = create_function_name(self._business["name"])
        self._file_generator.generate_script_file(
            f"workspace_{business_name}",
            [
                *import_code_lines,
                "",
                "",
                f"def workspace_{business_name}(shimoku_client: Client):",
                *["    " + line for line in main_code_lines],
                "",
            ],
        )
        # Create an __init__.py file for the imports to work
        self._file_generator.generate_script_file("__init__", [""])



import inspect
import json

from shimoku.playground.websockets_server import Subscription, define_event_method

from fastapi import FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
import uvicorn

import uuid
import datetime as dt
from typing import List, Optional, Any, Dict
from copy import copy

from starlette_graphene3 import GraphQLApp, make_graphiql_handler
import graphene

from dataclasses import asdict
from shimoku.playground import schema_classes
from shimoku.playground import schema_parameter_classses
from inspect import getmembers, isclass
import aiohttp

print_paths = True


def now_time_format():
    """
    Get the current time in the format that the API expects
    """
    return dt.datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def get_resource_name(_type, to_lower=True):
    """
    Get the resource name from the type
    :param _type: The type to get the resource name from
    :param to_lower: Whether to return the resource name with the first letter lowercased
    """
    cut = _type[:-7]
    if to_lower:
        cut = cut[0].lower() + cut[1:]
    # Ends with 'Exposed'
    return cut


def get_plural(parents_url_singular):
    """
    Get the plural of a parent url
    :param parents_url_singular: The singular of the parent url
    """
    if parents_url_singular in ["Data", "data", "business"]:
        return parents_url_singular
    elif parents_url_singular.endswith("s"):
        return parents_url_singular + "es"
    elif parents_url_singular.endswith("y") and parents_url_singular[-2] not in "aeiou":
        return parents_url_singular[:-1] + "ies"
    else:
        return parents_url_singular + "s"


def get_query_field(types: Dict[str, Any], db: Dict[str, Any], element_type: str):
    """
    Generate a graphql get query field for a given type
    :param types: a dictionary of all the types
    :param db: the database
    :param element_type: the type to generate the query field for
    """

    element_id_param = (
        get_resource_name(element_type) + "Id"
        if element_type not in ["AccountExposed", "UserExposed"]
        else "id"
    )

    def resolver(self, info, **kwargs):
        return db[element_type][kwargs[element_id_param]]

    return graphene.Field(
        types[element_type],
        **{element_id_param: graphene.ID(required=True)},
        resolver=resolver,
    )


def list_query_field(
    types: Dict[str, Any], db: Dict[str, Any], parent_type: str, element_type: str
):
    """
    Generate a graphql list query field for a given type
    :param types: a dictionary of all the types
    :param db: the database
    :param parent_type: the type of the parent
    :param element_type: the type to generate the query field for
    """

    parent_id_param = get_resource_name(parent_type) + "Id"
    query_params = {
        parent_id_param: graphene.ID(required=True),
        "limit": graphene.Int(),
        "from": graphene.Int(name="from"),
        "nextToken": graphene.String(name="nextToken"),
    }
    list_resolver_extra = None
    if hasattr(types[element_type], "list_resolver_extra"):
        list_resolver_extra = types[element_type].list_resolver_extra

        annotations = inspect.getfullargspec(list_resolver_extra).annotations

        for arg_name, arg_type in annotations.items():
            if arg_name not in ["args", "kwargs", "return"]:
                query_params[arg_name] = graphene.Argument(types[arg_type])

    def resolver(self, info, **kwargs):
        results = copy(
            getattr(
                db[parent_type][str(kwargs[parent_id_param])],
                get_plural(get_resource_name(element_type)),
            )
        )
        results.items = copy(results.items)
        if list_resolver_extra is not None:
            results.items = list_resolver_extra(items=results.items, **kwargs)
        total = len(results.items)
        offset = 0
        if kwargs.get("nextToken") is not None:
            next_token_index = int(kwargs["nextToken"])
            results.items = results.items[next_token_index:]
            offset = next_token_index
        elif "from" in kwargs:
            results.items = results.items[kwargs["from"] :]
            offset = kwargs["from"]
        if "limit" in kwargs:
            results.items = results.items[: kwargs["limit"]]
            if offset + kwargs["limit"] < total:
                results.nextToken = str(offset + kwargs["limit"])
        return results

    return graphene.Field(
        types[element_type + "List"], **query_params, resolver=resolver
    )


def generate_schema(fast_api_app, types, db, is_child_of):
    query_fields = {}
    for type_name, type_class in types.items():
        if (
            not type_name.endswith("Exposed")
            or "Filtered" in type_name
            or "Input" in type_name
        ):
            continue
        name = "get" + get_resource_name(type_name, to_lower=False)
        query_fields[name] = get_query_field(types, db, type_name)

        if type_name in is_child_of:
            parent_type = is_child_of[type_name]
            name = "list" + get_plural(get_resource_name(type_name, to_lower=False))
            query_fields[name] = list_query_field(types, db, parent_type, type_name)

    Query = type("Query", (graphene.ObjectType,), query_fields)

    # TODO: Implement mutations
    ##################-TEMPORAL!-#####################
    ####-just so frontend doesnt display an error-####
    ###########-fix when mutations needed-############
    from graphene import Mutation, InputObjectType

    class UpdateUserInputExposed(InputObjectType):
        id = graphene.ID(required=True)
        name = graphene.String()
        email = graphene.String()
        companyName = graphene.String()
        companyRole = graphene.String()
        createdAt = graphene.String()
        phoneNumber = graphene.String()
        birthDate = graphene.String()
        accountInUseId = graphene.String()
        shopifyMasterAccountId = graphene.String()
        cognitoId = graphene.String()
        updateApiToken = graphene.Boolean()
        kindOfUser = graphene.String()
        newsletterAccepted = graphene.String()

    class UpdateUser(Mutation):
        class Arguments:
            input = UpdateUserInputExposed(required=True)

        id = graphene.String()
        ok = graphene.Boolean()

        def mutate(self, info, input):
            return UpdateUser(ok=True, id=input.id)

    class MyMutations(graphene.ObjectType):
        updateUser = UpdateUser.Field()

    ##################################################
    ##################################################
    ##################################################

    schema = graphene.Schema(
        query=Query, subscription=Subscription, mutation=MyMutations
    )

    # Integrate with FastAPI (assuming you have a way to do this)
    graphql_app = GraphQLApp(schema, on_get=make_graphiql_handler())
    fast_api_app.add_route("/graphql", graphql_app)
    fast_api_app.add_websocket_route("/graphql", graphql_app)


def get_children(types: Dict[str, Any]):
    """
    Generate a dictionaries of all the children and parents of the exposed types
    :param types: a dictionary of all the types
    """
    is_child_of = {}
    is_parent_of = {}
    links = {}
    for type_name, type_class in types.items():
        if "Exposed" not in type_name or "Filtered" in type_name:
            continue
        for field_name, field_def in type_class.__dict__.items():
            if not hasattr(field_def, "_type"):
                continue
            field_type = field_def.type

            if not hasattr(field_type, "items"):
                element_type_name = field_type.__name__
                links[(type_name, field_name)] = element_type_name
            else:
                # is_child_of[element_type_name] = (type_name, field_name)
                element_type_name = field_type.items.of_type.__name__
                is_child_of[element_type_name] = type_name
                if type_name not in is_parent_of:
                    is_parent_of[type_name] = []
                is_parent_of[type_name].append(element_type_name)

    return is_child_of, is_parent_of


def clean_element(types: Dict[str, Any], element_type: str, element: Any):
    """
    Clean an element to be sent to the frontend or SDK
    :param types: a dictionary of all the types
    :param element_type: The type of the element
    :param element: The element to clean
    """
    dict_rep = {}
    for field_name, field_type in types[element_type].__dict__.items():
        if isinstance(
            field_type, (graphene.List, graphene.Field)
        ) or field_name.startswith("_"):
            continue
        first_upper = field_name[0].upper() + field_name[1:] + "Exposed"
        if first_upper in types:
            dict_rep[field_name + "Id"] = (
                getattr(element, field_name).id
                if hasattr(getattr(element, field_name), "id")
                else None
            )
        dict_rep[field_name] = getattr(element, field_name)

    return dict_rep


async def list_elements(
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    parent0Id: str,
    element_type: str,
) -> List[Dict[str, Any]]:
    """
    List all elements of a given type and parent id
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param parent0Id: The id of the parent element
    :param element_type: The type of the element
    """
    if parent0Id is not None:
        parent_type = is_child_of[element_type]
        if parent0Id not in db[parent_type]:
            raise HTTPException(
                status_code=404,
                detail=f"{get_resource_name(parent_type, to_lower=False)} with id {parent0Id} not found",
            )
        print(parent_type, element_type)
        parent0 = db[parent_type][parent0Id]
        if getattr(parent0, get_plural(get_resource_name(element_type))) is None:
            return []
        return [
            clean_element(types, element_type, copy(elm))
            for elm in getattr(
                parent0, get_plural(get_resource_name(element_type))
            ).items
        ]

    return [
        clean_element(types, element_type, copy(elm))
        for elm in db[element_type].values()
    ]


def set_parent(
    parent_type: str, parent0Id: str, input_class: type, element_type: str, params: dict
):
    """
    Set the parent of the element to be created
    :param parent_type: The type of the parent
    :param parent0Id: The id of the parent
    :param input_class: The class of the input
    :param element_type: The type of the element
    :param params: The parameters of the element
    """
    parent_id_key = f"{get_resource_name(parent_type)}Id"
    alternative_parent_id_key = (
        f"{get_resource_name(element_type)}{get_resource_name(parent_type, False)}Id"
    )
    if hasattr(input_class, parent_id_key):
        params[parent_id_key] = parent0Id
    elif hasattr(input_class, alternative_parent_id_key):
        params[alternative_parent_id_key] = parent0Id


def set_children_and_links(
    types: Dict[str, Any], db: Dict[str, Any], element_type: str, params: dict
):
    """
    Set the children and links of the element to be created
    :param types: a dictionary of all the types
    :param db: the database
    :param element_type: The type of the element
    :param params: The parameters of the element
    """
    for field_name, field_def in types[element_type].__dict__.items():
        if field_name.startswith("_"):
            continue
        if isinstance(field_def, graphene.Field) and hasattr(field_def.type, "items"):
            params[field_name] = field_def.type(items=[])
            if hasattr(params[field_name], "total"):
                params[field_name].total = 0
        elif field_name + "Id" in params and params[field_name + "Id"] is not None:
            field_name_id = field_name + "Id"
            link_type_name = field_name[0].upper() + field_name[1:]
            link_type_name = link_type_name + "Exposed"
            link_id = params[field_name_id]
            if link_type_name in types:
                if link_id not in db[link_type_name]:
                    raise HTTPException(
                        status_code=404,
                        detail=f"{link_type_name} with id {link_id} not found",
                    )
                if not hasattr(types[element_type], field_name_id):
                    del params[field_name_id]
                params[field_name] = db[link_type_name][link_id]
            else:
                raise HTTPException(
                    status_code=404, detail=f"{link_type_name} not found in types"
                )


def append_to_parent(
    db: Dict[str, Any],
    element_id: str,
    parent_type: str,
    parent0Id: str,
    element_type: str,
):
    """
    Append the element to the parent
    :param db: the database
    :param element_id: The id of the element
    :param parent_type: The type of the parent
    :param parent0Id: The id of the parent
    :param element_type: The type of the element
    """
    plural_name = get_plural(get_resource_name(element_type))
    list_field = getattr(db[parent_type][parent0Id], plural_name)
    if list_field is None:
        print("BAD")
    if list_field.items is None:
        print("BAD")
    list_field.items.append(db[element_type][element_id])
    if hasattr(list_field, "total"):
        list_field.total = (list_field.total if list_field.total else 0) + 1


def set_contained_classes(types: Dict[str, Any], element_type: str, params: dict):
    for k, v in params.items():
        if not hasattr(types[element_type], k):
            continue
        linked_type = getattr(types[element_type], k)
        if not hasattr(linked_type, "__name__"):
            continue
        type_of_contained_class_name = getattr(types[element_type], k).__name__
        if type_of_contained_class_name in types:
            if v is not None:
                params[k] = types[type_of_contained_class_name](**v)


async def create_element(
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    parent0Id: str,
    element_type: str,
    params: Any,
    input_class: type,
) -> Dict[str, Any]:
    """
    Create an element of type element_type
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param parent0Id: The id of the parent
    :param element_type: The type of the element
    :param params: The parameters of the element
    :param input_class: The class of the input
    """
    parent_type = is_child_of[element_type]
    if parent0Id not in db[parent_type]:
        raise HTTPException(
            status_code=404, detail=f"{element_type} with id {parent0Id} not found"
        )

    created_at = now_time_format()
    if not isinstance(params, dict):
        params = asdict(params)

    _id = str(uuid.uuid4())
    params.update({"id": _id})

    if hasattr(types[element_type], "createdAt"):
        params["createdAt"] = created_at

    set_contained_classes(types, element_type, params)
    set_parent(parent_type, parent0Id, input_class, element_type, params)
    set_children_and_links(types, db, element_type, params)

    if element_type not in db:
        db[element_type] = {}

    aux_params = {}
    for key, value in params.items():
        if hasattr(types[element_type], key):
            aux_params[key] = value
    params = aux_params

    db[element_type][_id] = types[element_type](**params)

    append_to_parent(db, _id, parent_type, parent0Id, element_type)

    return clean_element(types, element_type, db[element_type][_id])


async def update_element(
    types: Dict[str, Any], db: Dict[str, Any], element_type: str, Id: str, params: Any
):
    """
    Update an element of type element_type
    :param types: a dictionary of all the types
    :param db: the database
    :param element_type: The type of the element
    :param Id: The id of the element
    :param params: The parameters to update
    """
    if Id not in db[element_type]:
        raise HTTPException(
            status_code=404, detail=f"{element_type} with id {Id} not found"
        )
    params = asdict(params)
    for key, value in params.items():
        if value is not None:
            setattr(db[element_type][Id], key, value)

    return clean_element(types, element_type, db[element_type][Id])


async def delete_r(
    db: Dict[str, Any],
    is_parent_of: Dict[str, List[str]],
    delete_type: str,
    delete_id: str,
):
    """
    Delete an element of type delete_type and all its children
    :param db: the database
    :param is_parent_of: a dictionary of the children of the exposed types
    :param delete_type: The type of the element
    :param delete_id: The id of the element
    """
    if delete_id not in db[delete_type]:
        return
    if delete_type in is_parent_of:
        for child_type in is_parent_of[delete_type]:
            plural = get_plural(get_resource_name(child_type))
            plural = plural[0].lower() + plural[1:]
            children_to_delete = getattr(db[delete_type][delete_id], plural)
            if children_to_delete is None:
                continue
            for child in children_to_delete.items:
                if getattr(child, "id"):
                    await delete_r(db, is_parent_of, child_type, getattr(child, "id"))
    del db[delete_type][delete_id]


async def delete_element(
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    is_parent_of: Dict[str, List[str]],
    parent0Id: str,
    element_type: str,
    Id: str,
):
    """
    Delete an element of type element_type
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param is_parent_of: a dictionary of the children of the exposed types
    :param parent0Id: The id of the parent
    :param element_type: The type of the element
    :param Id: The id of the element
    """
    parent_type = is_child_of[element_type]
    plural_name = get_plural(get_resource_name(element_type))
    parent_children = getattr(db[parent_type][parent0Id], plural_name)
    parent_children.items = [
        elm
        for elm in getattr(db[parent_type][parent0Id], plural_name).items
        if elm.id != Id
    ]
    if hasattr(parent_children, "total"):
        parent_children.total = parent_children.total - 1
    if Id not in db[element_type]:
        raise HTTPException(
            status_code=404, detail=f"{element_type} with id {Id} not found"
        )
    await delete_r(db, is_parent_of, element_type, Id)
    return {"message": "Deleted"}


def define_get_method(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    parents_url: str,
    element_type: str,
):
    """
    Define the GET method for the element
    :param fast_api_app: The fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param parents_url: The url of the parent
    :param element_type: The type of the element
    """
    print("GET " + parents_url + "/{Id}") if print_paths else None

    @fast_api_app.get(parents_url + "/{Id}")
    async def get(Id: str):
        try:
            return clean_element(types, element_type, db[element_type][Id])
        except KeyError:
            raise HTTPException(
                status_code=404,
                detail=f'{element_type.replace("Exposed","").lower()} '
                f"with id {Id} not found",
            )


def define_list_method(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    parents_url: str,
    element_type: str,
):
    """
    Define the GET method for the list of elements
    :param fast_api_app: The fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param parents_url: The url of the parent
    :param element_type: The type of the element
    """
    print("GET " + get_plural(parents_url)) if print_paths else None

    @fast_api_app.get(get_plural(parents_url))
    async def list_elms(parent0Id: Optional[str] = None):
        return await list_elements(types, db, is_child_of, parent0Id, element_type)

    @fast_api_app.post(get_plural(parents_url))
    async def list_elms(parent0Id: Optional[str] = None):
        return await list_elements(types, db, is_child_of, parent0Id, element_type)


def define_create_method(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    parents_url: str,
    element_type: str,
    uppercase_type_name: str,
):
    """
    Define the POST method for the element if the input class exists
    :param fast_api_app: The fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param parents_url: The url of the parent
    :param element_type: The type of the element
    :param uppercase_type_name: The name of the type of the element with the first letter in uppercase
    """
    create_input_name = "Create" + uppercase_type_name + "Input"
    if create_input_name in types:
        input_class = types[create_input_name]
        print("POST " + parents_url) if print_paths else None

        @fast_api_app.post(parents_url)
        async def _create(parent0Id: Optional[str], params: input_class):
            return await create_element(
                types, db, is_child_of, parent0Id, element_type, params, input_class
            )


def define_batch_create_method(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    parents_url: str,
    element_type: str,
    uppercase_type_name: str,
):
    """
    Define the POST method for the batch creation of elements if the input class exists
    :param fast_api_app: The fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param parents_url: The url of the parent
    :param element_type: The type of the element
    :param uppercase_type_name: The name of the type of the element with the first letter in uppercase
    """
    batch_create_input_name = "BatchCreate" + uppercase_type_name + "Input"
    if batch_create_input_name in types:
        print("POST " + parents_url + "/batch") if print_paths else None

        @fast_api_app.post(parents_url + "/batch")
        async def batch_create(parent0Id: Optional[str], request: Request):
            items = await request.json()
            if not isinstance(items, list):
                raise HTTPException(
                    status_code=400,
                    detail=f"Expected a list of {batch_create_input_name}",
                )
            # TODO: Get a more generic name for the input class
            for item in items:
                await create_element(
                    types,
                    db,
                    is_child_of,
                    parent0Id,
                    element_type,
                    item,
                    types["CreateDataInput"],
                )
            return {"result": "ok"}


def define_update_method(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    parents_url: str,
    element_type: str,
    uppercase_type_name: str,
):
    """
    Define the PATCH method for the element if the input class exists
    :param fast_api_app: The fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param parents_url: The url of the parent
    :param element_type: The type of the element
    :param uppercase_type_name: The name of the type of the element with the first letter in uppercase
    """
    update_input_name = "Update" + uppercase_type_name + "Input"
    if update_input_name in types:
        input_class = types[update_input_name]
        print("PATCH " + parents_url + "/{Id}") if print_paths else None

        @fast_api_app.patch(parents_url + "/{Id}")
        async def update(Id: str, params: input_class):
            return await update_element(types, db, element_type, Id, params)


def define_delete_method(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, Any],
    is_parent_of: Dict[str, Any],
    parents_url: str,
    element_type: str,
    uppercase_type_name: str,
):
    """
    Define the DELETE method for the element if the input class exists
    :param fast_api_app: The fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param is_parent_of: a dictionary of the children of the exposed types
    :param parents_url: The url of the parent
    :param element_type: The type of the element
    :param uppercase_type_name: The name of the type of the element with the first letter in uppercase
    """
    delete_input_name = "Delete" + uppercase_type_name + "Input"
    if delete_input_name in types:
        print("DELETE " + parents_url + "/{Id}") if print_paths else None

        @fast_api_app.delete(parents_url + "/{Id}")
        async def delete(parent0Id: str, Id: str):
            return await delete_element(
                db, is_child_of, is_parent_of, parent0Id, element_type, Id
            )


def generate_rest_crud_methods(
    fast_api_app: FastAPI,
    types: Dict[str, Any],
    db: Dict[str, Any],
    is_child_of: Dict[str, str],
    is_parent_of: Dict[str, List[str]],
    element_type_name: str,
):
    """
    Generate the CRUD methods for the element
    :param fast_api_app: the fast api app
    :param types: a dictionary of all the types
    :param db: the database
    :param is_child_of: a dictionary of the parents of the exposed types
    :param is_parent_of: a dictionary of the children of the exposed types
    :param element_type_name: The type of the element
    """
    base_url = "/external/v1/"

    aux_type = element_type_name
    parents_url = ""
    parent_index = 0
    while aux_type in is_child_of:
        parent_type = is_child_of[aux_type]
        parents_url = (
            get_resource_name(parent_type)
            + "/{parent"
            + str(parent_index)
            + "Id}/"
            + parents_url
        )
        aux_type = parent_type
        parent_index += 1

    element_type = types[element_type_name]
    parents_url = (
        base_url
        + parents_url
        + (
            get_resource_name(element_type_name)
            if not hasattr(element_type, "rest_name")
            else element_type.rest_name()
        )
    )

    uppercased_type_name = element_type_name[:-7]

    print(element_type_name)

    define_get_method(fast_api_app, types, db, parents_url, element_type_name)
    define_list_method(
        fast_api_app, types, db, is_child_of, parents_url, element_type_name
    )
    define_create_method(
        fast_api_app,
        types,
        db,
        is_child_of,
        parents_url,
        element_type_name,
        uppercased_type_name,
    )
    define_batch_create_method(
        fast_api_app,
        types,
        db,
        is_child_of,
        parents_url,
        element_type_name,
        uppercased_type_name,
    )
    define_update_method(
        fast_api_app, types, db, parents_url, element_type_name, uppercased_type_name
    )
    define_delete_method(
        fast_api_app,
        types,
        db,
        is_child_of,
        is_parent_of,
        parents_url,
        element_type_name,
        uppercased_type_name,
    )


def define_webhook_methods(
    fast_api_app: FastAPI, types: Dict[str, Any], db: Dict[str, Any]
):
    """
    Define the methods for the webhooks, these cant be created with the generate_crud_methods function
    because the pattern is different
    :param fast_api_app: the fast api app
    :param types: a dictionary of all the types
    :param db: the database
    """
    CreateWebhookInput = types["CreateActivityWebHookInput"]

    @fast_api_app.post(
        "/external/v1/universe/{parent3Id}/business/{parent2Id}/app/{parent1Id}/"
        "activity/{parent0Id}/webhook"
    )
    async def create(parent0Id: str, params: CreateWebhookInput):
        webhook_class = types["ActivityWebHookExposed"]
        params = asdict(params)
        params.update(
            {
                "id": str(uuid.uuid4()),
                "activity": db["ActivityExposed"][parent0Id],
                "createdAt": now_time_format(),
                "updatedAt": now_time_format(),
                "activityId": parent0Id,
            }
        )
        db["ActivityExposed"][parent0Id].webhook = webhook_class(**params)
        return clean_element(
            types, "ActivityWebHookExposed", db["ActivityExposed"][parent0Id].webhook
        )

    @fast_api_app.post(
        "/external/v1/universe/{parent4Id}/business/{parent3Id}/app/{parent2Id}/activity/{parent1Id}/"
        "run/{parent0Id}/triggerWebhook"
    )
    async def call_webhook(
        parent4Id: str, parent3Id: str, parent2Id: str, parent1Id: str, parent0Id: str
    ):
        if db["ActivityExposed"][parent1Id].webhook is None:
            raise HTTPException(status_code=404, detail="Webhook not found")

        webhook = db["ActivityExposed"][parent1Id].webhook
        headers = webhook.headers
        if headers is None:
            headers = {}
        else:
            headers = json.loads(headers)
        body = {
            "runId": parent0Id,
            "activityId": parent1Id,
            "appId": parent2Id,
            "businessId": parent3Id,
            "universeId": parent4Id,
        }
        url = webhook.url
        method = webhook.method

        db["RunExposed"][parent0Id].isWebhookCalled = True

        if db["RunExposed"][parent0Id].logs is None:
            db["RunExposed"][parent0Id].logs = []

        db["RunExposed"][parent0Id].logs.items.append(
            types["LogExposed"](
                id=str(uuid.uuid4()),
                dateTime=now_time_format(),
                message="Process started",
                runId=parent0Id,
                run=db["RunExposed"][parent0Id],
                severity="INFO",
            )
        )

        # async call to webhook with aiohttp
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method, url, headers=headers, json=json.dumps(body)
            ) as response:
                await response.text()

        return {"STATUS": "OK"}


def create_api() -> FastAPI:
    """
    Create the API with FastAPI and Strawberry
    """
    types = {c[0]: c[1] for c in getmembers(schema_classes, isclass)}
    types.update({c[0]: c[1] for c in getmembers(schema_parameter_classses, isclass)})

    # Add the fields that cannot be added in the class definition
    for type_name, type_class in types.items():
        if hasattr(type_class, "add_fields"):
            types[type_name] = type_class.add_fields()

    active_universe_plan = types["ActiveUniversePlan"](
        id="local",
        planType=types["PlanType"](id="local", limits=None, price=None, type=None),
    )
    db = {
        "UserExposed": {
            "userPlaygroundId": types["UserExposed"](id="userPlaygroundId")
        },
        "UniverseFilteredExposed": {
            "local": types["UniverseFilteredExposed"](
                id="local",
                name="local",
                activeUniversePlanId="local",
                activeUniversePlan=active_universe_plan,
                actions=types["ActionExposedList"](items=[]),
                universeApiKeys=types["UniverseApiKeyExposedList"](items=[]),
            )
        },
        "UniverseExposed": {
            "local": types["UniverseExposed"](
                id="local",
                name="local",
                activeUniversePlanId="local",
                activeUniversePlan=active_universe_plan,
                actions=types["ActionExposedList"](items=[]),
                universeApiKeys=types["UniverseApiKeyExposedList"](items=[]),
            )
        },
        "BusinessExposed": {
            "local": types["BusinessExposed"](
                id="local", name="local", type="local", createdAt="2021-01-01"
            ),
        },
    }
    db["UniverseExposed"]["local"].business = types["BusinessExposedList"](
        items=[db["BusinessExposed"]["local"]]
    )
    db["UniverseFilteredExposed"]["local"].business = types["BusinessExposedList"](
        items=[db["BusinessExposed"]["local"]]
    )
    db["UniverseExposed"]["local"].modules = types["ModuleExposedList"](items=[])
    db["UniverseFilteredExposed"]["local"].modules = types["ModuleExposedList"](
        items=[]
    )
    db["BusinessExposed"]["local"].apps = types["AppExposedList"](items=[], total=0)
    db["BusinessExposed"]["local"].dashboards = types["DashboardExposedList"](
        items=[], total=0
    )
    db["BusinessExposed"]["local"].rolePermissions = types["RolePermissionExposedList"](
        items=[]
    )
    db["BusinessExposed"]["local"].modules = types["ModuleExposedList"](items=[])
    db["BusinessExposed"]["local"].universe = db["UniverseFilteredExposed"]["local"]
    db["AccountExposed"] = {
        "local": types["AccountExposed"](
            id="local", business=db["BusinessExposed"]["local"]
        )
    }
    db["UserExposed"]["userPlaygroundId"].accounts = types["AccountExposedList"](
        items=[db["AccountExposed"]["local"]]
    )

    fast_api_app = FastAPI()

    is_child_of, is_parent_of = get_children(types)

    generate_schema(fast_api_app, types, db, is_child_of)

    # Generate the CRUD methods for each type and define the webhook methods
    for _type in types:
        if _type in is_child_of:
            generate_rest_crud_methods(
                fast_api_app, types, db, is_child_of, is_parent_of, _type
            )

    define_webhook_methods(fast_api_app, types, db)
    define_event_method(fast_api_app)

    return fast_api_app


# Create the API
app = create_api()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def main(server_host: Optional[str] = None, server_port: Optional[int] = None):
    if not server_host:
        server_host = "127.0.0.1"
    if not server_port:
        server_port = 8000
    uvicorn.run(app, host=server_host, port=server_port)


if __name__ == "__main__":
    main()



from shimoku import Client
from .data.table_with_labels import table_with_labels_data
import numpy as np


def table_with_labels(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "table-test-with-labels")
    shimoku_client.plt.table(
        data=table_with_labels_data,
        order=0,
        rows_size=10,
        page_size=100,
        initial_sort_column="date",
        categorical_columns=["filtA", "filtB"],
        label_columns={
            ("x", "outlined"): {
                3: [0, 255, 0],
                4: "#42F548",
                5: "#666666",
                6: "#4287f5",
                7: [255, 0, 0],
            },
            ("y", "outlined"): {
                (0, 25): "red",
                (25, 50): "orange",
                (50, 75): "yellow",
                (75, np.inf): "green",
            },
            "z": {
                (0, 25): [100, 100, 100],
                (25, 50): "yellow",
                (50, 75): "orange",
                (75, np.inf): "red",
            },
            "a": {
                (0, 25): [100, 100, 100],
                (25, 50): "yellow",
                (50, 75): "orange",
                (75, np.inf): "red",
            },
            "filtA": "#666666",
            ("filtB", "outlined"): [125, 54, 200],
            "name": "warning",
            "name2": {
                "Ana": "active",
                "Laura": "error",
                "Audrey": "warning",
                "Jose": "caution",
                "Jorge": "main",
            },
            ("name3", "outlined"): "neutral",
        },
    )



from typing import Optional, TYPE_CHECKING

import asyncio

from shimoku.api.base_resource import Resource, ResourceCache
from shimoku.api.resources.role import (
    Role,
    create_role,
    get_role,
    get_roles,
    delete_role,
)
from shimoku.api.resources.app import App

from dataclasses import dataclass

if TYPE_CHECKING:
    from shimoku.api.resources.business import Business
from shimoku.exceptions import BoardError
import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


class Dashboard(Resource):
    """Dashboard resource class"""

    resource_type = "dashboard"
    alias_field = "name"
    plural = "dashboards"

    class AppDashboard(Resource):
        """AppDashboard resource class"""

        resource_type = "appDashboard"
        plural = "appDashboards"

        def __init__(
            self,
            parent: "Dashboard",
            uuid: Optional[str] = None,
            app: App = None,
            db_resource: Optional[dict] = None,
        ):
            params = dict(
                appId=app["id"] if app else None,
            )
            super().__init__(
                parent=parent,
                uuid=uuid,
                db_resource=db_resource,
                check_params_before_creation=["appId"],
                params=params,
            )

        async def delete(self):
            """Deletes the appDashboard"""
            return await self._base_resource.delete()

    @dataclass
    class PublicPermission:
        """PublicPermission class"""

        isPublic: bool
        permission: str
        token: str

    def __init__(
        self,
        parent: "Business",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = dict(
            name=alias,
            order=0,
            isDisabled=False,
            theme={},
            publicPermission=self.PublicPermission(
                isPublic=False, permission="READ", token="default token"
            ),
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            children=[Dashboard.AppDashboard, Role],
            check_params_before_creation=["name"],
            params=params,
            params_to_serialize=["theme"],
        )

        self.currently_in_use = False

    async def delete(self):
        """Deletes the dashboard"""
        if self.currently_in_use:
            log_error(
                logger,
                f"Workspace {str(self)} is currently in use and cannot be deleted",
                BoardError,
            )
        return await self._base_resource.delete()

    async def update(self):
        """Updates the dashboard"""
        return await self._base_resource.update()

    async def list_app_ids(self) -> list[str]:
        """Returns the list of app ids in the dashboard"""
        cache: ResourceCache = self._base_resource.children[Dashboard.AppDashboard]
        cache_list = await cache.list()
        return list(set([app_dashboard["appId"] for app_dashboard in cache_list]))

    async def insert_app(self, app: App):
        """Inserts an app in the dashboard
        :param app: The App to insert
        """
        dashboards = await self._base_resource.parent.get_dashboards()
        app_ids_lists = await asyncio.gather(
            *[dashboard.list_app_ids() for dashboard in dashboards]
        )
        app_ids = [app_id for app_ids_list in app_ids_lists for app_id in app_ids_list]
        if app["id"] in app_ids:
            logger.warning(
                f"Menu path {str(app)} already exists in another board, it will appear in both boards"
            )
        await self._base_resource.create_child(Dashboard.AppDashboard, appId=app["id"])
        logger.info(f"Menu path {str(app)} added to board {str(self)}")

    async def remove_app(self, app: App):
        """Removes an app from the dashboard
        :param app: The App to remove
        """
        tasks = []
        cache: ResourceCache = self._base_resource.children[Dashboard.AppDashboard]
        cache_list = await cache.list()
        for app_dashboard in cache_list:
            if app_dashboard["appId"] == app["id"]:
                tasks.append(cache.delete(uuid=app_dashboard["id"]))

        await asyncio.gather(*tasks)
        logger.info(f"Menu path {str(app)} removed from board {str(self)}")

    async def remove_all_apps(self):
        """Removes all apps from the dashboard"""
        tasks = []
        cache: ResourceCache = self._base_resource.children[Dashboard.AppDashboard]
        cache_list = await cache.list()
        for app_dashboard in cache_list:
            tasks.append(cache.delete(uuid=app_dashboard["id"]))

        await asyncio.gather(*tasks)

    # Role methods
    get_role = get_role
    get_roles = get_roles
    create_role = create_role
    delete_role = delete_role



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )
    from shimoku.api.resources.report import Report


async def code_gen_from_filter(self: "AppCodeGen", report: "Report") -> list[str]:
    """Generate code for a filter report.
    :param report: report to generate code from
    :return: list of code lines
    """
    filter_def = report["properties"]["filter"][0]
    mapping = report["properties"]["mapping"][0]
    field_name = filter_def["field"]
    data_set = await self._app.get_data_set(mapping["id"])
    field = field_name if data_set["columns"] else mapping[field_name]
    report_params = [
        f'    order={report["order"]},',
        f'    data="{data_set["name"]}",',
        f'    field="{field}",',
    ]
    if report["sizeColumns"] != 4:
        report_params.append(f'    cols_size={report["sizeColumns"]},')
    if report["sizeRows"] != 1:
        report_params.append(f'    rows_size={report["sizeRows"]},')
    if report["sizePadding"] != "0,0,0,0":
        report_params.append(f'    padding="{report["sizePadding"]}",')
    if filter_def["inputType"] == "CATEGORICAL_MULTI":
        report_params.append("    multi_select=True,")
    return ["shimoku_client.plt.filter(", *report_params, ")"]



import asyncio
import graphene
import queue
import uuid
from pydantic import BaseModel
import warnings

warnings.filterwarnings("ignore", module="pydantic")
import strawberry

warnings.filterwarnings("default", module="pydantic")
from shimoku.utils import EventType
from fastapi import FastAPI


EventTypeGRAPHENE = graphene.Enum.from_enum(EventType)


class Event(graphene.ObjectType):
    id = graphene.String()
    type = graphene.Field(EventTypeGRAPHENE)
    resourceId = graphene.String()
    content = graphene.String()
    universeId = graphene.String()


events_queue: queue.Queue = queue.Queue()
new_event_available: asyncio.Event = asyncio.Event()


class Subscription(graphene.ObjectType):
    onEventCreated = graphene.Field(Event)

    async def subscribe_onEventCreated(root, info):
        global new_event_available
        new_event_available = asyncio.Event()
        while True:
            await new_event_available.wait()  # Wait until a new event is signaled
            while not events_queue.empty():  # Process all events in the queue
                yield events_queue.get()
            new_event_available.clear()  # Reset the event


@strawberry.type
class CreateEventInput(BaseModel):
    type: str
    resourceId: str
    content: str


def define_event_method(fast_api_app: FastAPI):
    @fast_api_app.post("/external/v1/universe/{parent1Id}/business/{parent0Id}/event")
    async def create(parent1Id: str, parent0Id: str, params: CreateEventInput):
        event = Event(
            id=str(uuid.uuid4()),
            type=params.type,
            resourceId=params.resourceId,
            content=params.content,
            universeId=parent1Id,
        )
        events_queue.put(event)
        new_event_available.set()  # Signal that a new event is available
        return event



from shimoku import Client
from .data.flow_and_rainfall import flow, rainfall, time


def rainfall_line(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "rainfall-line")
    how_many = 100

    data = []
    for i in range(0, len(flow), max(3, len(flow) // how_many)):
        data.append(
            {
                "Date": time[i],
                "flow": flow[i],
                "flow+1": flow[(i + 101) % len(flow)],
                "flow+2": flow[(i + 202) % len(flow)],
                "rainfall": rainfall[i],
                "rainfall+1": rainfall[(i + 101) % len(flow)],
                "rainfall+2": rainfall[(i + 202) % len(flow)],
            }
        )

    shimoku_client.plt.top_bottom_line(
        data=data,
        order=0,
        x="Date",
        top_names=["flow", "flow+1", "flow+2"],
        bottom_names=["rainfall", "rainfall+1", "rainfall+2"],
        title="rainfall and flow",
        x_axis_name="Date",
        top_axis_name="flow(m³/s)",
        bottom_axis_name="rainfall(mm)",
    )

    data = []
    for i in range(0, len(flow), max(1, len(flow) // how_many)):
        data.append({"Date": time[i], "flow": flow[i], "rainfall": rainfall[i]})

    shimoku_client.plt.top_bottom_line(
        data=data,
        order=1,
        x="Date",
        top_names=["flow"],
        bottom_names=["rainfall"],
        title="rainfall and flow",
        x_axis_name="Date",
        top_axis_name="flow(m³/s)",
        bottom_axis_name="rainfall(mm)",
    )



from ...report import Report


class Button(Report):
    report_type = "BUTTON"

    default_properties = dict(
        **Report.default_properties, text="More Info", align="stretch", events={}
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional


class ActivityTemplatesLayerHeader:
    """
    This class is used to interact with the API at the activity template metadata level.
    """

    def get_activity_template(
        self,
        uuid: Optional[str] = None,
        name_version: Optional[tuple[str, str]] = None,
    ) -> dict:
        """Get a workspace
        :param name_version: name and version of the activity template
        :param uuid: UUID of the activity template
        """
        pass



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
    InvertibleDictFunction,
    generic_echart_function_params,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


from shimoku.utils import get_arg_names

common_opt = get_common_echart_options()


def get_pie_chart_options_function() -> InvertibleDictFunction:
    pie_func = InvertibleDictFunction(pie_chart)
    pie_func.set_items(get_common_echart_options())
    pie_func.delete_items(["yAxis", "xAxis", "grid"])
    pie_func.delete_items(offset=["toolbox", "feature"], keys=["magicType"])
    pie_func.set_items(
        final=True,
        offset=["series", 0],
        items={
            "type": "pie",
            "data": "#set_data#",
            "itemStyle": {"borderWidth": 0, "borderRadius": 5},
            "radius": "70%",
        },
    )
    pie_func.custom_operation(
        func=lambda params: params.update(
            {
                "fields": [
                    {
                        "name": pie_func.params["names"],
                        "value": pie_func.params["values"],
                    }
                ]
            }
        ),
        inverse_func=lambda params: (
            params.update(
                {
                    "names": pie_func.params["fields"][0]["name"],
                    "values": pie_func.params["fields"][0]["value"],
                }
            ),
            params.pop("fields"),
        ),
    )
    pie_func.params_to_params(generic_echart_function_params, replace=True)
    return pie_func


async def pie_chart(
    self: "PlotLayer",
    order: int,
    names: str,
    values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None,
    rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a pie chart"""
    local_vars = locals()
    kwargs = {
        k: local_vars[k]
        for k in get_arg_names(pie_chart)
        if k != "self" and k in local_vars
    }
    pie_func = get_pie_chart_options_function()
    await self._create_echart(**pie_func(**kwargs))


async def doughnut_chart(
    self: "PlotLayer",
    order: int,
    names: str,
    values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None,
    rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a doughnut chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["yAxis"]
    del common_options["xAxis"]
    del common_options["grid"]
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["tooltip"]["trigger"] = "item"
    series_options.update(
        {
            "type": "pie",
            "data": "#set_data#",
            "itemStyle": {"borderWidth": 0, "borderRadius": 5},
            "radius": ["40%", "70%"],
            "label": {"show": False, "position": "center"},
            "emphasis": {
                "label": {"show": True, "fontSize": "30", "fontWeight": "bold"}
            },
            "labelLine": {"show": False},
        }
    )
    common_options["series"] = [series_options]

    await self._create_echart(
        fields=[{"name": names, "value": values}],
        options=common_options,
        data_mapping_to_tuples=await self._choose_data(order, data),
        option_modifications=option_modifications,
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )


async def rose_chart(
    self: "PlotLayer",
    order: int,
    names: str,
    values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None,
    rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a rose chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["yAxis"]
    del common_options["xAxis"]
    del common_options["grid"]
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["tooltip"]["trigger"] = "item"
    series_options.update(
        {
            "type": "pie",
            "roseType": "area",
            "data": "#set_data#",
            "itemStyle": {"borderWidth": 0, "borderRadius": 5},
            "radius": ["10%", "70%"],
        }
    )
    common_options["series"] = [series_options]

    await self._create_echart(
        fields=[{"name": names, "value": values}],
        options=common_options,
        data_mapping_to_tuples=await self._choose_data(order, data),
        option_modifications=option_modifications,
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



from shimoku import Client
from .data.horizontal_bar import horizontal_bar_data


def horizontal_bar_chart(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "horizontal-bar-test")

    shimoku_client.plt.horizontal_bar(data=horizontal_bar_data, x="name", order=0)

    shimoku_client.plt.horizontal_bar(
        data=horizontal_bar_data,
        x="name",
        x_axis_name="Axis x",
        y_axis_name="Axis y",
        title="Title",
        order=1,
        rows_size=3,
        cols_size=10,
        padding="0,0,0,1",
    )



import pandas as pd
from os import path
from shimoku import Client

data_path = path.join(path.dirname(path.abspath(__file__)), "data")


def shimoku_gauges(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "shimoku-gauges")
    df = pd.read_csv(path.join(data_path, "test_stack_distribution.csv"))

    value_columns = [col for col in df.columns if col != "Segment"]
    df = df[["Segment"] + value_columns]

    gauges_data = pd.DataFrame(columns=["name", "value", "color"])
    df_transposed = df.transpose().reset_index().drop(0)
    value_columns = [col for col in df_transposed.columns if col != "index"]
    gauges_data["value"] = df_transposed[value_columns].apply(
        lambda row: sum(row), axis=1
    )
    gauges_data["name"] = df_transposed["index"]
    gauges_data["color"] = range(1, len(df_transposed) + 1)

    order = shimoku_client.plt.shimoku_gauges_group(
        gauges_data=gauges_data,
        order=0,
        cols_size=12,
        rows_size=3,
        calculate_percentages=True,
    )

    shimoku_client.plt.shimoku_gauge(value=-60, order=order, color=1)

    order += 1
    shimoku_client.plt.shimoku_gauge(
        value=60, order=order, name="test", color="status-error"
    )

    order += 1
    shimoku_client.plt.shimoku_gauge(
        value=-90, order=order, name="test", color="#FF0000"
    )



from shimoku import Client
from .data.annotation_chart import annotated_data, annotated_data_1, annotated_data_2


def annotation_chart(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "Annotation Chart")

    shimoku_client.plt.annotated_chart(data=[annotated_data], x="date", y="x", order=0)

    shimoku_client.plt.annotated_chart(
        order=7,
        x="date",
        y=["Síntoma [1]", "Síntoma [2]"],
        annotations="Annotation",
        data=[annotated_data_1, annotated_data_2],
        slider_config={"max": 100, "defaultValue": 50},
        slider_marks=[("Low", 15), ("Medium", 50), ("High", 85)],
    )



from __future__ import absolute_import

from typing import Dict, Optional

from shimoku.async_execution_pool import (
    AutoAsyncExecutionPool,
    decorate_class_to_auto_async,
)

from shimoku.api.client import ApiClient
from shimoku.api.user_access_classes.universes_layer import UniversesLayer, Universe
from shimoku.api.user_access_classes.generated_headers.UniversesLayerHeader import (
    UniversesLayerHeader,
)
from shimoku.api.user_access_classes.businesses_layer import WorkspacesLayer, Business
from shimoku.api.user_access_classes.generated_headers.WorkspacesLayerHeader import (
    WorkspacesLayerHeader,
)
from shimoku.api.user_access_classes.activity_templates_layer import (
    ActivityTemplatesLayer,
)
from shimoku.api.user_access_classes.generated_headers.ActivityTemplatesLayerHeader import (
    ActivityTemplatesLayerHeader,
)
from shimoku.api.user_access_classes.dashboards_layer import BoardsLayer, Dashboard
from shimoku.api.user_access_classes.generated_headers.BoardsLayerHeader import (
    BoardsLayerHeader,
)
from shimoku.api.user_access_classes.apps_layer import MenuPathsLayer, App
from shimoku.api.user_access_classes.generated_headers.MenuPathsLayerHeader import (
    MenuPathsLayerHeader,
)
from shimoku.api.user_access_classes.reports_layer import ComponentsLayer
from shimoku.api.user_access_classes.generated_headers.ComponentsLayerHeader import (
    ComponentsLayerHeader,
)
from shimoku.api.user_access_classes.data_sets_layer import DataSetsLayer
from shimoku.api.user_access_classes.generated_headers.DataSetsLayerHeader import (
    DataSetsLayerHeader,
)
from shimoku.api.user_access_classes.files_layer import FilesLayer
from shimoku.api.user_access_classes.generated_headers.FilesLayerHeader import (
    FilesLayerHeader,
)
from shimoku.api.user_access_classes.activities_layer import ActivitiesLayer
from shimoku.api.user_access_classes.generated_headers.ActivitiesLayerHeader import (
    ActivitiesLayerHeader,
)
from shimoku.api.user_access_classes.actions_layer import ActionsLayer
from shimoku.api.user_access_classes.generated_headers.ActionsLayerHeader import (
    ActionsLayerHeader,
)
from shimoku.plt.plt_layer import PlotLayer
from shimoku.plt.generated_headers.PlotLayerHeader import PlotLayerHeader
from shimoku.plt.utils import create_normalized_name

from shimoku.ai.ai_layer import AILayer
from shimoku.ai.generated_headers.AILayerHeader import AILayerHeader

from shimoku.utils import IN_BROWSER, EventType

from shimoku.exceptions import WorkspaceError, BoardError

import shimoku_components_catalog.html_components

import logging
from shimoku.execution_logger import (
    log_error,
    configure_logging,
    logging_before_and_after,
)

logger = logging.getLogger(__name__)


class Client:
    def __init__(
        self,
        universe_id: str = "local",
        environment: str = "production",
        access_token: Optional[str] = None,
        config: Optional[Dict] = None,
        verbosity: str = None,
        async_execution: bool = False,
        local_port: int = 8000,
        retry_attempts: int = 5,
    ):
        self.playground: bool = universe_id == "local" and not access_token
        if self.playground:
            access_token = "local"
        if universe_id == "local" and not self.playground:
            log_error(
                logger,
                "Local universe can only be used in playground mode.",
                AttributeError,
            )
        self.access_token = access_token
        self.environment = environment
        self._access_token = access_token
        self.universe_id = universe_id
        self.workspace_id = None
        self.board_id = None
        self.menu_path_id = None
        if not config:
            config = {}

        self.server_host = "127.0.0.1"
        self.local_port = local_port

        self.configure_logging = configure_logging
        if verbosity:
            self.configure_logging(verbosity)

        if access_token and access_token != "":
            config = {"access_token": access_token}

        self._api_client = ApiClient(
            config=config,
            environment=environment,
            playground=self.playground,
            server_host=self.server_host,
            server_port=local_port,
            retry_attempts=retry_attempts,
        )
        self._universe_object = Universe(self._api_client, uuid=universe_id)
        self._business_object: Optional[Business] = None
        self._app_object: Optional[App] = None
        self._dashboard_object: Optional[Dashboard] = None

        self._async_pool = AutoAsyncExecutionPool(api_client=self._api_client)

        if async_execution:
            self.activate_async_execution()
        else:
            self.activate_sequential_execution()

        self.universes: UniversesLayerHeader = decorate_class_to_auto_async(
            UniversesLayer, self._async_pool
        )(self._api_client)
        self.workspaces: WorkspacesLayerHeader = decorate_class_to_auto_async(
            WorkspacesLayer, self._async_pool
        )(self._universe_object)
        self.activity_templates: ActivityTemplatesLayerHeader = (
            decorate_class_to_auto_async(ActivityTemplatesLayer, self._async_pool)(
                self._universe_object
            )
        )
        self.actions: ActionsLayerHeader = decorate_class_to_auto_async(
            ActionsLayer, self._async_pool
        )(self._universe_object)
        self.boards: BoardsLayerHeader = decorate_class_to_auto_async(
            BoardsLayer, self._async_pool
        )(self._business_object)
        self.menu_paths: MenuPathsLayerHeader = decorate_class_to_auto_async(
            MenuPathsLayer, self._async_pool
        )(self._business_object)
        self.components: ComponentsLayerHeader = decorate_class_to_auto_async(
            ComponentsLayer, self._async_pool
        )(self._app_object)
        self.data: DataSetsLayerHeader = decorate_class_to_auto_async(
            DataSetsLayer, self._async_pool
        )(self._app_object)
        self.io: FilesLayerHeader = decorate_class_to_auto_async(
            FilesLayer, self._async_pool
        )(self._app_object)
        self.activities: ActivitiesLayerHeader = decorate_class_to_auto_async(
            ActivitiesLayer, self._async_pool
        )(self._app_object)
        self.ai: AILayerHeader = decorate_class_to_auto_async(
            AILayer, self._async_pool
        )(self.access_token, self._universe_object, self._app_object)

        self._reuse_data_sets = False
        self._shared_dfs = {}
        self._shared_custom_data = {}
        self.plt: PlotLayerHeader = decorate_class_to_auto_async(
            PlotLayer, self._async_pool
        )(self._async_pool, self._app_object, self._reuse_data_sets)

        self._async_pool.current_app = self._app_object
        self._async_pool.universe = self._universe_object

        self.html_components = shimoku_components_catalog.html_components

    def set_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> None:
        """Set workspace id for the client.
        :param uuid: Workspace uuid
        :param name: Workspace name
        """
        if self._api_client.playground:
            uuid, name = "local", None

        @logging_before_and_after(logging_level=logger.info, name="set_workspace")
        async def a_set_workspace(_self):
            if self._business_object:
                self._business_object.currently_in_use = False
            business: Optional[Business] = await self._universe_object.get_business(
                uuid=uuid, name=name
            )
            if not business:
                log_error(
                    logger,
                    f"Workspace {name if name else uuid} not found.",
                    WorkspaceError,
                )
            self._business_object = business
            self._business_object.currently_in_use = True
            self.workspace_id = self._business_object["id"]

            if self._app_object:
                self.plt.raise_if_cant_change_path()
                self._app_object.currently_in_use = False
            self._app_object = None
            self.menu_path_id = None
            if self._dashboard_object:
                self._dashboard_object.currently_in_use = False
            self._dashboard_object = None
            self.board_id = None

            self.boards.__init__(self._business_object)
            self.menu_paths.__init__(self._business_object)
            self.components.__init__(self._app_object)
            self.activities.__init__(self._app_object)
            self.plt.__init__(self._async_pool, self._app_object, self._reuse_data_sets)
            self.data.__init__(self._app_object)
            self.io.__init__(self._app_object)
            self.ai.__init__(self.access_token, self._universe_object, self._app_object)

        return self._async_pool.auto_async_func_call(
            name="set_workspace",
            func_self=self,
            func=a_set_workspace,
        )

    def pop_out_of_menu_path(self) -> None:
        """Pop out of the menu path."""

        @logging_before_and_after(
            logging_level=logger.info, name="pop_out_of_menu_path"
        )
        async def a_pop_out_of_menu_path(_self):
            data_names = self.plt.get_shared_data_names()
            if data_names:
                logger.info(
                    f"Shared data entries will no longer be available: {data_names}, set them again if needed."
                )
            self.plt.clear_context()
            self._app_object.currently_in_use = False
            self._app_object = None
            self.menu_path_id = None

        return self._async_pool.auto_async_func_call(
            name="pop_out_of_menu_path",
            func_self=self,
            func=a_pop_out_of_menu_path,
        )

    def set_board(self, name: str) -> None:
        """Set the board in use for the following apps being called.
        :param name: board name
        """
        if not self._business_object:
            log_error(
                logger,
                "Workspace not set. Please use set_workspace() method first.",
                AttributeError,
            )
        if self._app_object:
            self.pop_out_of_menu_path()
        if self._dashboard_object:
            self._dashboard_object.currently_in_use = False

        @logging_before_and_after(logging_level=logger.info, name="set_board")
        async def a_set_board(_self):
            self._dashboard_object = await self._business_object.get_dashboard(
                name=name
            )
            self._dashboard_object.currently_in_use = True
            self.board_id = self._dashboard_object["id"]

        return self._async_pool.auto_async_func_call(
            name="set_board",
            func_self=self,
            func=a_set_board,
        )

    def pop_out_of_board(self) -> None:
        """Pop out of the dashboard."""
        if not self._business_object:
            log_error(
                logger,
                "Workspace not set. Please use set_workspace() method first.",
                AttributeError,
            )
        if not self._dashboard_object:
            log_error(
                logger,
                "Board not set. Please use set_board() method first.",
                BoardError,
            )
        if self._app_object:
            self.pop_out_of_menu_path()

        @logging_before_and_after(logging_level=logger.info, name="pop_out_of_board")
        async def a_pop_out_of_board(_self):
            self._dashboard_object.currently_in_use = False
            self._dashboard_object = None
            self.board_id = None

        return self._async_pool.auto_async_func_call(
            name="pop_out_of_board",
            func_self=self,
            func=a_pop_out_of_board,
        )

    def enable_caching(self):
        """Enable caching."""
        self._api_client.cache_enabled = True

    def disable_caching(self):
        """Disable caching."""
        self._api_client.cache_enabled = False

    def reuse_data_sets(self):
        """Reuse data sets from the api."""
        self._reuse_data_sets = True
        if self._app_object:
            self.plt.reuse_data_sets = True

    def update_data_sets(self):
        """Update data sets in the api."""
        self._reuse_data_sets = False
        if self._app_object:
            self.plt.reuse_data_sets = False

    async def _change_app(self, menu_path: str, dont_add_to_dashboard: bool):
        """Change app in use for the following calls.
        :param menu_path: Menu path of the app
        :param dont_add_to_dashboard: Whether to add the menu path to the dashboard
        """
        if self._app_object:
            self._app_object.currently_in_use = False

        app: App = await self._business_object.get_app(name=menu_path)
        self._app_object = app
        app.currently_in_use = True
        self.activities.__init__(self._app_object)
        self.components.__init__(self._app_object)
        self.plt.__init__(self._async_pool, self._app_object, self._reuse_data_sets)
        self.data.__init__(self._app_object)
        self.io.__init__(self._app_object)
        self.ai.__init__(self.access_token, self._universe_object, self._app_object)

        self.menu_path_id = self._app_object["id"]

        if dont_add_to_dashboard:
            return

        if not self._dashboard_object:
            self._dashboard_object = await self._business_object.get_dashboard(
                name="Default Name"
            )

        if self._app_object["id"] not in await self._dashboard_object.list_app_ids():
            await self._dashboard_object.insert_app(self._app_object)

        self.menu_path_id = self._app_object["id"]

    def set_menu_path(
        self,
        name: str,
        sub_path: Optional[str] = None,
        dont_add_to_dashboard: bool = False,
    ) -> None:
        """Set menu path for the client.
        :param name: Menu path
        :param sub_path: Sub path
        :param dont_add_to_dashboard: Whether to add the menu path to the dashboard
        """
        if not self._business_object:
            log_error(
                logger,
                "Workspace not set. Please use set_workspace() method first.",
                AttributeError,
            )
        normalized_name = create_normalized_name(name)
        path = sub_path if sub_path else None
        data_names = []
        if self._app_object:
            self.plt.raise_if_cant_change_path()
            if self._app_object["normalizedName"] == normalized_name:
                self.plt.change_path(path)
                return
            data_names = self.plt.get_shared_data_names()

        @logging_before_and_after(logging_level=logger.info, name="set_menu_path")
        async def a_set_menu_path(_self):
            await self._change_app(name, dont_add_to_dashboard)
            self.plt.change_path(path)
            if data_names:
                logger.info(
                    f"Shared data entries will no longer be available: {data_names}, set them again if needed."
                )

        return self._async_pool.auto_async_func_call(
            name="set_menu_path",
            func_self=self,
            func=a_set_menu_path,
        )

    async def _create_business_contents_updated_event(self):
        """Create a business contents updated event."""
        if not self._business_object:
            return
        await self._business_object.create_event(
            EventType.BUSINESS_CONTENTS_UPDATED, {}, self._business_object["id"]
        )
        self.plt.clear_events_for_components()
        logger.info("Business contents updated event created")

    def run(self) -> None:
        """Run the tasks in the execution pool."""

        async def a_run(_self):
            self._async_pool.ending_tasks[
                "Business_contents_updated"
            ] = self._create_business_contents_updated_event()

        return self._async_pool.auto_async_func_call(
            name="run",
            func_self=self,
            func=a_run,
        )

    def activate_async_execution(self):
        """Activate async execution of the tasks."""
        self._async_pool.sequential = False

    def activate_sequential_execution(self):
        """Activate sequential execution of the tasks."""
        self._async_pool.sequential = True

    def get_api_calls_counter(self):
        """Get the number of api calls made."""
        return self._api_client.call_counter

    def __getattribute__(self, item):
        """Get attribute of the client."""
        if False and not IN_BROWSER:
            if item in ["boards", "menu_paths"] and not self._business_object:
                log_error(
                    logger,
                    "Workspace not set. Please use set_workspace() method first.",
                    AttributeError,
                )
            if (
                item in ["activities", "plt", "components", "data", "io"]
                and not self._app_object
            ):
                log_error(
                    logger,
                    "Menu path not set. Please use set_menu_path() method first.",
                    AttributeError,
                )
        return object.__getattribute__(self, item)



import requests

# Download the data
res = requests.get(
    url="https://echarts.apache.org/examples/data/asset/data/confidence-band.json"
)

# Convert it to json
confidence_data = res.json()

for dat in confidence_data:
    dat["value"] = dat["value"] * 100
    dat["l"] = dat["l"] * 100
    dat["u"] = dat["u"] * 100



from shimoku import Client


def heatmap(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "heatmap")

    shimoku_client.plt.heatmap(
        data="heatmap data",
        x="xAxis",
        y="yAxis",
        values="value",
        order=0,
        calculate_color_range=True,
        variant="minimal",
    )

    shimoku_client.plt.heatmap(
        data="heatmap data",
        x="xAxis",
        y="yAxis",
        values="value",
        order=1,
        continuous=True,
        title="Heatmap Chart",
        rows_size=2,
        cols_size=12,
        x_axis_name="xAxis",
        y_axis_name="yAxis",
    )



import argparse
import subprocess
from abc import ABC
from typing import Optional
import inspect
import asyncio

from aiohttp.client_exceptions import ClientConnectorError

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion

from prompt_toolkit.history import FileHistory

from shimoku.utils import get_args_with_defaults

from shimoku.execution_logger import configure_logging
from shimoku.cli.utils import (
    INTERACTIVE_HISTORY_PATH,
    get_current_profile,
    chose_from_available_paths,
    tab_main_navigation_key_binding,
)

from typing import Iterable
import shlex

import logging

logger = logging.getLogger(__name__)


class CLICustomCompleter(Completer):
    USEFUL_VARIABLES = [
        "environment",
        "access_token",
        "universe_id",
        "workspace_id",
        "local_port",
        "menu_path",
        "board",
        "activity",
    ]

    def __init__(self, parser_dict: dict, variables: dict):
        # parser_dict = {
        #     'name': '',
        #     'commands': [
        #         {...},
        #         ...
        #     ],
        #     'args': [
        #         '--arg_name',
        #         ...
        #     ]
        # }
        self.parser_dict = parser_dict
        self.variables = variables

    def get_completions(self, document, complete_event) -> Iterable[Completion]:
        text = document.text_before_cursor
        if text.startswith("$ "):
            used_vars = [
                var
                for var in self.USEFUL_VARIABLES
                if var.startswith(text[2:]) and var not in self.variables
            ]
            for var in used_vars:
                yield Completion(var, start_position=-len(text[2:]))
            return

        last_char_is_space = text[-1] == " " if text else False
        if "  " in text:
            return []
        words = text.split(" ")

        completion_scope = self.parser_dict
        used_args = []
        next_word_is_arg_value = False
        last_word = "" if not words else words[-1]

        for word in words:
            if next_word_is_arg_value:
                if word in completion_scope["args"]:
                    return []
                if last_word == word:
                    break
                next_word_is_arg_value = False
            elif word in [command["name"] for command in completion_scope["commands"]]:
                completion_scope = [
                    command
                    for command in completion_scope["commands"]
                    if command["name"] == word
                ][0]
                used_args = []
            elif word in completion_scope["args"]:
                next_word_is_arg_value = True
                used_args.append(word)
            elif word and last_char_is_space:
                return []

        if next_word_is_arg_value:
            for var in self.variables:
                if var.startswith(last_word):
                    yield Completion(var, start_position=-len(last_word))
            return []

        for command in completion_scope["commands"]:
            if last_char_is_space or command["name"].startswith(last_word):
                yield Completion(command["name"], start_position=-len(last_word))
        for arg in completion_scope["args"]:
            if arg not in used_args and (
                last_char_is_space or arg.startswith(last_word)
            ):
                yield Completion(arg, start_position=-len(last_word))


class CLIFuncParam:
    def __init__(
        self,
        name: Optional[str] = None,
        arg_type: Optional[type] = None,
        arg_help: Optional[str] = None,
        mandatory: bool = True,
        default: Optional = None,
        prompt: bool = False,
        choices: Optional[list] = None,
        real_name: Optional[str] = None,
        action: Optional[str] = None,
        group: Optional[str] = None,
        alt_name: Optional[str] = None,
        is_path: bool = False,
    ):
        self.name = name
        self.arg_type = arg_type
        self.arg_help = arg_help
        self.mandatory = mandatory
        self.default = default
        self.prompt = prompt
        self.choices = choices
        self.real_name = real_name
        self.action = action
        self.group = group
        self.alt_name = alt_name
        self.is_path = is_path

        if default is not None:
            self.mandatory = False
            if arg_type and not isinstance(default, arg_type):
                raise TypeError(
                    f"The default value {default} is not of type {arg_type}"
                )
            if choices and default not in choices:
                raise ValueError(
                    f"The default value {default} is not in the choices {choices}"
                )

        if prompt or group:
            self.mandatory = False
            if default is not None:
                raise ValueError(
                    "Do not set the default value of the CliFuncParam when prompt or group is set to True"
                )


def normalize_name_for_cli(name: str):
    return name.replace("_", "-").lower()


class CLIParser(ABC):
    def __init__(
        self,
        name: str = "shimoku",
        description: str = "This is the Shimoku cli, it is used to interact with the Shimoku services",
        func: callable = None,
        parent: Optional["CLIParser"] = None,
        arguments: Optional[list[CLIFuncParam]] = None,
    ):
        self.name = name
        self.description = description

        self.command_names: list[str] = []
        self.commands: dict[str, CLIParser] = {}
        self.arguments: list[CLIFuncParam] = []
        self.argument_groups = {}
        self.argument_groups_var_names = {}
        self.shell_commands_enabled = False

        if not parent:
            self.parser = argparse.ArgumentParser(
                self.name, description=self.description
            )
        else:
            parent.commands[name] = self
            self.parser = parent.subparsers.add_parser(
                self.name, description=self.description, help=self.description
            )

        self.subparsers = self.parser.add_subparsers(
            dest="command", help="Available commands"
        )

        if arguments:
            for argument in arguments:
                self.add_argument(argument)
        if func:
            self.parser.set_defaults(func=self._decorator_for_self_func(func))
        else:
            self.parser.set_defaults(func=self._default_func)

    def _default_func(self):
        self.parser.print_help()

    def _decorator_for_self_func(self, func: callable):
        def prompt_arg(argument: CLIFuncParam, kwargs: dict):
            if argument.group:
                for arg_name in self.argument_groups_var_names[argument.group]:
                    if kwargs[arg_name] is not None:
                        raise ValueError(
                            f"You can not use the argument {argument.name} with the argument {arg_name}"
                        )
            arg_type = (
                argument.arg_type
                if "Optional" not in str(argument.arg_type)
                else argument.arg_type.__args__[0]
            )
            input_name = argument.real_name if argument.real_name else argument.name
            if not argument.is_path:
                arg_inp = input(f"{input_name}: ")
            else:
                print(f"Select the path for the argument {input_name}:")
                arg_inp = chose_from_available_paths()

            kwargs[argument.name] = arg_type(arg_inp) if arg_inp else None

            if argument.choices and kwargs[argument.name] not in argument.choices:
                raise ValueError(
                    f"The value {kwargs[argument.name]} is not in the choices {argument.choices}"
                )

        def wrapper(**kwargs):
            for argument in self.arguments:
                if argument.prompt and kwargs.get(argument.real_name) is None:
                    prompt_arg(argument, kwargs)
                if argument.real_name and argument.name in kwargs:
                    kwargs[argument.real_name] = kwargs.pop(argument.name)

            return func(**kwargs)

        return wrapper

    def decor_add_command(
        self, name: Optional[str] = None, common_args: list[CLIFuncParam] = None
    ):
        common_args = common_args or []

        def decorator(func: callable):
            func_params = list(inspect.signature(func).parameters.keys())
            func_type_params = {
                param.name: param.annotation
                for param in inspect.signature(func).parameters.values()
            }
            for func_param in func_params:
                if func_param not in func_type_params:
                    func_type_params[func_param] = str
            comment = str(func.__doc__)
            c_help = "\n".join(
                [line for line in comment.split("\n") if ":param" not in line]
            ).strip()
            params_helps = [line for line in comment.split("\n") if ":param" in line]
            params_helps = [line.split(":param ")[1] for line in params_helps]
            params_helps = {
                line.split(":")[0].strip(): line.split(":")[1].strip()
                for line in params_helps
            }
            default_args = get_args_with_defaults(func)
            func_args = []
            for func_param in func_params:
                if func_param not in params_helps:
                    continue
                dafault_value = default_args.get(func_param)
                if isinstance(dafault_value, CLIFuncParam):
                    func_param_obj = dafault_value
                    if func_param_obj.name is None:
                        func_param_obj.name = normalize_name_for_cli(func_param)
                    if (
                        func_param_obj.arg_type is None
                        and func_param_obj.action is None
                    ):
                        func_param_obj.arg_type = func_type_params[func_param]
                    if func_param_obj.arg_help is None:
                        func_param_obj.arg_help = params_helps[func_param]
                    if func_param_obj.real_name is None:
                        func_param_obj.real_name = func_param
                    else:
                        raise ValueError(
                            "Do not set the real_name of the CliFuncParam, it is set automatically"
                        )
                else:
                    arg_type = func_type_params[func_param]
                    mandatory = True
                    if "Optional" in str(arg_type):
                        arg_type = arg_type.__args__[0]
                        mandatory = False

                    func_param_obj = CLIFuncParam(
                        name=normalize_name_for_cli(func_param),
                        arg_type=arg_type,
                        arg_help=params_helps[func_param],
                        mandatory=mandatory,
                        default=dafault_value,
                    )
                if func_param_obj.name in func_args:
                    raise ValueError(
                        f"The argument {func_param_obj.name} is already registered"
                    )
                func_args.append(func_param_obj)

            self.add_command(
                name=normalize_name_for_cli(name if name else func.__name__),
                description=c_help,
                func=func,
                arguments=func_args + common_args,
            )
            return func

        return decorator

    def add_command(
        self,
        name: str = None,
        description: str = None,
        func: callable = None,
        arguments: Optional[list[CLIFuncParam]] = None,
    ) -> "CLIParser":
        command = CLIParser(
            name=name,
            description=description,
            func=func,
            parent=self,
            arguments=arguments,
        )
        if command.name in self.command_names:
            raise ValueError(f"The command {command.name} is already registered")
        self.command_names.append(command.name)
        return command

    def add_argument(self, argument: CLIFuncParam):
        def add_prefix(name: str):
            return ("-" if len(name) == 1 else "--") + name

        if argument.name in self.arguments:
            raise ValueError(f"The argument {argument.name} is already registered")
        params = {
            k: v
            for k, v in dict(
                action=argument.action,
                type=argument.arg_type,
                help=argument.arg_help,
                required=argument.mandatory,
                default=argument.default,
                choices=argument.choices,
            ).items()
            if v is not None
        }
        argument_names = [add_prefix(argument.name)]
        if argument.alt_name:
            argument_names.append(add_prefix(argument.alt_name))

        if argument.group:
            if argument.group not in self.argument_groups:
                self.argument_groups[
                    argument.group
                ] = self.parser.add_mutually_exclusive_group()
                self.argument_groups_var_names[argument.group] = []
            self.argument_groups[argument.group].add_argument(*argument_names, **params)
            self.argument_groups_var_names[argument.group].append(argument.name)
        else:
            self.parser.add_argument(*argument_names, **params)
        self.arguments.append(argument)

    @staticmethod
    async def execute_args(args, user_vars: Optional[dict] = None):
        args_ = vars(args).copy()
        args_.pop("interactive", None)
        args_.pop("shell_commands_enabled", None)
        args_.pop("command", None)
        args_.pop("func", None)

        if user_vars:
            user_vars = {
                k.replace("$", "").replace("-", "_"): v for k, v in user_vars.items()
            }
            for func_param in args_.keys():
                if args_[func_param] is None and func_param in user_vars:
                    args_[func_param] = user_vars.get(func_param)
                    logger.info(
                        f"Using saved value for unspecified argument {func_param}"
                    )

        result = args.func(**args_)
        if inspect.isawaitable(result):
            await result

    async def parse_args(self):
        # argcomplete.autocomplete(self.parser)
        try:
            args = self.parser.parse_args()
        except SystemExit:
            return
        if hasattr(args, "interactive") and args.interactive:
            self.shell_commands_enabled = args.shell_commands_enabled
            await self.parse_args_interactive()
            return
        await self.execute_args(args)

    def gather_commands_and_options(self):
        commands_and_options = {"name": self.name, "commands": [], "args": []}
        for command in self.commands.values():
            commands_and_options["commands"].append(
                command.gather_commands_and_options()
            )
        for argument in self.arguments:
            if argument.name in ["interactive", "shell-commands-enabled"]:
                continue
            commands_and_options["args"].append("--" + argument.name)
        commands_and_options["args"].extend(["--help", "-h"])
        return commands_and_options

    async def parse_args_interactive(self):
        # Dictionary to store user-defined variables
        variables = {}

        # Gather possible commands and options
        commands_and_options = self.gather_commands_and_options()
        completer = CLICustomCompleter(commands_and_options, variables)

        # Trim the history file to 1000 lines
        with open(INTERACTIVE_HISTORY_PATH, "r+") as file:
            lines = file.readlines()
            trimmed_lines = lines[-1000:]
            file.seek(0)
            file.truncate()
            file.writelines(trimmed_lines)

        session = PromptSession(
            completer=completer,
            key_bindings=tab_main_navigation_key_binding(add_space=True),
            history=FileHistory(INTERACTIVE_HISTORY_PATH),
        )
        configure_logging("INFO")

        while True:
            try:
                text = await session.prompt_async(
                    f"S-{get_current_profile().upper()}> "
                )
                if not text:
                    continue

                # Check for variable assignment
                if text.startswith("$"):
                    try:
                        print()
                        # Splitting the input while respecting quoted sections
                        parts = shlex.split(text)
                        for i in range(1, len(parts), 2):
                            var_name, var_value = parts[i], parts[i + 1]
                            variables["$" + var_name] = var_value
                            print(f"Variable {var_name} set to {var_value}")
                            print()
                    except Exception as e:
                        if isinstance(e, (ValueError, IndexError)):
                            print("Invalid variable assignment syntax.")
                            print()
                        else:
                            raise e
                    continue
                if text.startswith("unset$"):
                    _, *var_names = text.split(" ")
                    for var_name in var_names:
                        if "$" + var_name not in variables:
                            print(f"Variable {var_name} not set")
                            print()
                            continue
                        variables.pop("$" + var_name)
                        print(f"Variable {var_name} unset")
                        print()
                    continue

                # Replace variables in user input
                for var_name, var_value in variables.items():
                    text = text.replace(var_name, var_value)

                if "$" in text:
                    print(
                        "Warning: It's possible that not all variables were replaced, please check your input."
                    )
                    print()
                if text in ["exit", "esc"]:
                    break
            except asyncio.CancelledError:
                continue
            except KeyboardInterrupt:
                break
            except EOFError:
                break

            if text.startswith("show "):
                text = text.replace("show ", "")
                print(text)
                print()
                continue
            try:
                args = self.parser.parse_args(text.rstrip().split(" "))
                await self.execute_args(args, variables)
            except SystemExit as e:
                if not self.shell_commands_enabled or e.code == 0:
                    continue
                print()
                confirmation = input(
                    f'The command: "{text}" \n'
                    f"could not be interpreted by the Shimoku parser\n"
                    "do you want to execute it as a shell command? (y/n): "
                )
                if confirmation.lower() == "y":
                    print()
                    try:
                        subprocess.run(text.split(), check=True)
                    except FileNotFoundError:
                        print(f"Command not found: {text}")
                    except subprocess.CalledProcessError:
                        print(f"Command failed: {text}")
            except Exception as e:
                if isinstance(
                    e,
                    (
                        ValueError,
                        TypeError,
                        KeyError,
                        AttributeError,
                        RuntimeError,
                        IndexError,
                        ClientConnectorError,
                    ),
                ):
                    logger.error(e)
